# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import numpy as np
import logging; logger = logging.getLogger(__name__)
from opendrift.models.oceandrift import OceanDrift, Lagrangian3DArray


class PlastElement(Lagrangian3DArray):
    variables = Lagrangian3DArray.add_variables([
        ('terminal_velocity', {'dtype': np.float32,
                               'units': 'm/s',
                               'level': OceanDrift.CONFIG_LEVEL_ESSENTIAL,
            'description': 'Positive value means rising particles (positive buoyancy)',
                               'default': 0.01})])


class PlastDrift(OceanDrift):
    """Trajectory model based on the OpenDrift framework.

    Propagation of plastics particles with ocean currents and
    additional Stokes drift and wind drag.

    Developed at MET Norway.

    """

    ElementType = PlastElement

    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        'sea_surface_wave_stokes_drift_x_velocity': {'fallback': 0},
        'sea_surface_wave_stokes_drift_y_velocity': {'fallback': 0},
        'sea_surface_wave_significant_height': {'fallback': 0},
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'ocean_vertical_diffusivity': {'fallback': 0.02, 'profiles': True},
        'ocean_mixed_layer_thickness': {'fallback': 50},
        'sea_floor_depth_below_sea_level': {'fallback': 10000},
        'land_binary_mask': {'fallback': None},
        }


    def __init__(self, *args, **kwargs):

        # Call parent constructor
        super(PlastDrift, self).__init__(*args, **kwargs)

        # Add specific config settings
        self._add_config({
            # TODO: this option should be moved to OceanDrift
            'vertical_mixing:mixingmodel': {'type': 'enum',
                'enum': ['randomwalk', 'analytical'], 'default': 'analytical',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description':
                    'Scheme to be used for vertical turbulent mixing'},
            })

        self._set_config_default('drift:vertical_mixing', True)
        self._set_config_default('drift:vertical_advection', True)
        self._set_config_default('drift:use_tabularised_stokes_drift', True)
        self._set_config_default('general:coastline_action', 'previous')
        self._set_config_default('vertical_mixing:diffusivitymodel', 'windspeed_Sundby1983')

    def update(self):
        """Update positions and properties of elements."""

        # Simply move particles with ambient current
        self.advect_ocean_current()

        self.update_particle_depth()

        # Advect particles due to Stokes drift
        self.stokes_drift()

        # Advect particles due to wind-induced shear near surface
        self.advect_wind()

    def update_particle_depth(self):

        if self.get_config('drift:vertical_mixing') is True:

            if self.get_config('vertical_mixing:mixingmodel') == 'randomwalk':
                logger.debug('Turbulent mixing of particles using random walk')
                self.vertical_mixing()

            if self.get_config('vertical_mixing:mixingmodel') == 'analytical':
                logger.debug('Submerging according to wind')
                self.elements.z = -np.random.exponential(
                    scale=self.environment.ocean_vertical_diffusivity/
                            self.elements.terminal_velocity,
                    size=self.num_elements_active())




"""
Class that represents the product type

With validation

Also maintains the products types and labels mapping
"""

from pathlib import Path
import csv

from ...util.many_many import ManyMany
from ..common.validators import EnumValidator
from .validation.warnings import WARNINGS


class TypeLabelsMap(ManyMany):
    """
    class to maintain a many to many relationship between product types and labels

    The ``.product_types`` attribute is a mapping with the labels as keys,
    and product types as values.

    The ``.labels`` attribute is a mapping with the product type as keys,
    and the associated labels as values.
    """

    product_types = ManyMany.right
    labels = ManyMany.left


def load_from_csv_file(filepath=None):
    """
    Loads the product types and labels mapping from a CSV file

    :param filepath=None: The name of the file to load from. If not
                          provided, it will look for:
                          "product_types_and_labels.csv" next to this module

    """

    if filepath is None:
        filepath = Path(__file__).parent / "product_types_and_labels.csv"

    with open(filepath, newline='', encoding="utf-8") as csvfile:
        # skip the header:
        while True:
            line = csvfile.readline()
            if line.strip().startswith("DATA TABLE:"):
                break

        ptypes_labels = {}
        reader = csv.reader(csvfile, dialect='excel')
        labels = next(reader)[2:]
        for row in reader:
            pt = row[0]
            for i, val in enumerate(row[2:]):
                if val.strip():
                    ptypes_labels.setdefault(pt, set()).add(labels[i])
            # add the pt to the labels
            # not doing this anymore.
            # ptypes_labels.setdefault(pt, set()).add(pt)
        return ptypes_labels


PRODUCT_TYPE_LABEL_MAPPING = load_from_csv_file()
PRODUCT_TYPES = tuple(PRODUCT_TYPE_LABEL_MAPPING)

types_to_labels = TypeLabelsMap(PRODUCT_TYPE_LABEL_MAPPING)

# I'd much rather not hard-code this, but how else to do it?
# this is used by the validation code
DOESNT_NEED_API = set(('Refined Product NOS',
                       'Refinery Intermediate',
                       'Solvent',
                       'Natural Plant Oil',
                       'Other'))


class ProductType(str):
    _valid_types = PRODUCT_TYPES
    _validator = EnumValidator(_valid_types,
                               WARNINGS["W003"],
                               case_insensitive=True)

    @classmethod
    def validate(cls, value):
        if not value:
            return [WARNINGS["W002"]]
        return cls._validator(value)










ERRORS = {
    # E01* : meta data related
    "E010": "Record has no oil_id: every record must have an ID",
    "E011": ("Record has invalid oil_id: every record must have a valid ID. "
             "{} is not valid"),
    "E012": "Reference year: {} is not a valid year (between {} and {})",
    "E013": "Review Status: {} is not valid, must be one of {}",
    # E03* -- physical properties related
    "E030": "Oils must have an API",
    "E031": "No Properties data at all",
    "E032": 'Distillation type is "{}", it must be one of: {}',

    # E04* -- units or values
    "E040": "Value for {}: {} is out of range: unit error?",
    "E041": "Value for {}: {} must be between 0 and 1",
    "E042": "Must have a value for {}",
    "E043": "API, {} does not match density at 60F. API should be: {:.1f}",

    # E05* -- duplicates, etc
    "E050": "Duplicate {} in {}",
}

ERRORS = {code: (code + ": " + msg) for code, msg in ERRORS.items()}



from types import LambdaType
from ..basereader import BaseReader

def none_or_cmp(a, b, cmp):
    if a is None:
        return b
    if b is None:
        return a

    return cmp(a, b)

class Combined(BaseReader):
    """
    A combination of two readers.
    """
    a: BaseReader
    b: BaseReader
    op: LambdaType

    def __init__(self, a, b, op):
        self.a = a
        self.b = b
        self.op = op

        self.variables = list(set(self.a.variables).intersection(self.b.variables))

        self.start_time = none_or_cmp(self.a.start_time, self.b.start_time, max)
        self.end_time = none_or_cmp(self.a.end_time, self.b.end_time, min)

        self.xmin = -180
        self.xmax = 180
        self.ymin = -90
        self.ymax = 90

        self.name = f'Combined({a.name} | {b.name})'

        self.proj4 = '+proj=latlong'

        super().__init__()

    def covers_positions(self, lon, lat):
        return np.intersect1d(self.a.covers_positions(lon, lat), self.b.covers_positions(lon, lat))

    def covers_time(self, time):
        return self.a.covers_time(time) and self.b.covers_time(time)

    def get_variables_interpolated(self, variables, *args, **kwargs):
        assert set(variables).issubset(self.variables), f"{variables} is not subset of {self.variables}"

        env_a, env_profiles_a = self.a.get_variables_interpolated(variables, *args, **kwargs)
        env_b, env_profiles_b = self.b.get_variables_interpolated(variables, *args, **kwargs)

        variables = [
            var for var in env_a.keys() if var not in ['x', 'y', 'z', 'time']
        ]

        for var in variables:
            env_a[var] = self.op(env_a[var], env_b[var])

        if env_profiles_a is not None:
            variables = [
                var for var in env_profiles_a.keys() if var not in ['x', 'y', 'z', 'time']
            ]
            for var in variables:
                env_profiles_a[var] = self.op(env_profiles_a[var], env_profiles_b[var])

        return env_a, env_profiles_a




"""
ManyMany

class for managing a many to many relationship in a pair of dicts

The two mappings are "left" and "right"

The left one is originally populated by the initializer.

After that they are equivalent -- when you add something to one the other is updated.

Currently there is no way to remove anything

YOu can access copies of the mappings with:

``ManyMany.left``
and
``ManyMany.right``

They are copies, so that mutating them won't break the internal data.

"""

from collections import defaultdict

import copy

class ManyMany:

    def __init__(self, initial_data=None):
        """
        initialize a ManyMany structure

        :param initial_data: initial data for the left dict. of the form:
                             {key1: iterable_of_values,
                              key2: iterable_of_values,
                              ...
                              }

        all values must be hashable
        """

        self._left_dict = {key: set(values) for key, values in initial_data.items()}
        self._rebuild_right()

        self._lefthash = self._dict_hash(self._left_dict)
        self._righthash = self._dict_hash(self._right_dict)

    def _rebuild_right(self):
        """
        rebuilds the right dict to match the left
        """
        self._right_dict = self._rebuild(self._left_dict)

    def _rebuild_left(self):
        """
        rebuilds the left dict to match the right
        """
        self._left_dict = self._rebuild(self._right_dict)

    @staticmethod
    def _rebuild(source):
        """
        builds a "reversed" dict from a source dict
        """
        new = {}
        for key, values in source.items():
            for val in values:
                new.setdefault(val, set()).add(key)
        return new

    @staticmethod
    def _dict_hash(d):
        """
        Provide a hash of a dict with a sequence of hashable items as the values

        This was to be used to know if an internal dict was changed,
          but it turns out that's not useful (at least not in a robust way)
        """
        hashable = frozenset({k: frozenset(v) for k, v in d.items()}.items())
        return hash(hashable)

    def add_to_left(self, key, value):
        """
        add a new value to the left dict

        :param key: the key the value is to be added to
        :param value: the value to be added

        If the key is already there, the value will be added to the corresponding set

        A new key and set will be created if it is not already there.
        """
        self._left_dict.setdefault(key, set()).add(value)
        self._rebuild_right()

    def add_to_right(self, key, value):
        """
        add a new value to the right dict

        :param key: the key the value is to be added to
        :param value: the value to be added

        If the key is already there, the value will be added to the corresponding set

        A new key and set will be created if it is not already there.
        """
        self._right_dict.setdefault(key, set()).add(value)
        self._rebuild_left()

    @property
    def left(self):
        """
        A copy of the left dict

        It's a copy, so it won't change the internal ones if mutated
        """
        return copy.deepcopy(self._left_dict)

    @property
    def right(self):
        """
        A copy of the right dict

        It's a copy, so it won't change the internal ones if mutated
        """
        return copy.deepcopy(self._right_dict)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import sys
from datetime import datetime, timedelta
import numpy as np
from scipy.interpolate import interp1d
import logging; logger = logging.getLogger(__name__)
from opendrift.models.basemodel import OpenDriftSimulation
from opendrift.elements import LagrangianArray
from opendrift.models.physics_methods import verticaldiffusivity_Large1994, verticaldiffusivity_Sundby1983, gls_tke, skillscore_liu_weissberg

# Defining the oil element properties
class Lagrangian3DArray(LagrangianArray):
    """Extending LagrangianArray for elements moving in 3 dimensions
    The Particle may be buoyant and/or subject to vertical mixing
    buoyant bahaviour is described by terminal velocity
    """

    variables = LagrangianArray.add_variables([
        ('wind_drift_factor', {'dtype': np.float32,
                               'units': '1',
            'description': 'Elements at surface are moved with this '
                'fraction of the vind vector, in addition to currents '
                'and Stokes drift',
                               'default': 0.02}),
        ('current_drift_factor', {'dtype': np.float32,
                                  'units': '1',
            'description': 'Elements are moved with this fraction of the '
                            'current vector, in addition to currents '
                            'and Stokes drift',
                               'default': 1}),
        ('terminal_velocity', {'dtype': np.float32,
                               'units': 'm/s',
            'description': 'Terminal rise/sinking velocity (buoyancy) '
                'in the ocean column',
                               'default': 0.})])


class OceanDrift(OpenDriftSimulation):
    """Open source buoyant particle trajectory model based on OpenDrift.

        Developed at MET Norway

        Generic module for particles that move in 3 dimensions
        and may be to vertical turbulent mixing
        with the possibility for positive or negative buoyancy

        Particles could be e.g. oil droplets, plankton, nutrients or sediments,
        Model may be subclassed for more specific behaviour.

    """

    ElementType = Lagrangian3DArray

    max_speed = 1  # m/s

    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'upward_sea_water_velocity': {'fallback': 0},
        'ocean_vertical_diffusivity': {'fallback': 0,
             'profiles': True},
        'sea_surface_wave_significant_height': {'fallback': 0},
        'sea_surface_wave_stokes_drift_x_velocity': {'fallback': 0},
        'sea_surface_wave_stokes_drift_y_velocity': {'fallback': 0},
        'sea_surface_wave_period_at_variance_spectral_density_maximum':
            {'fallback': 0},
        'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment': {'fallback': 0},
        'surface_downward_x_stress': {'fallback': 0},
        'surface_downward_y_stress': {'fallback': 0},
        'turbulent_kinetic_energy': {'fallback': 0},
        'turbulent_generic_length_scale': {'fallback': 0},
        'ocean_mixed_layer_thickness': {'fallback': 50},
        'sea_floor_depth_below_sea_level': {'fallback': 10000},
        'land_binary_mask': {'fallback': None},
        }

    # The depth range (in m) which profiles shall cover
    required_profiles_z_range = [-20, 0]

    def __init__(self, *args, **kwargs):

        if 'machine_learning_dict' in kwargs:
            logger.debug('Machine learning correction supplied.')
            mld = kwargs['machine_learning_dict']
            del kwargs['machine_learning_dict']
            import pickle
            from tensorflow import keras
            from tools.normalizations import generate_one_normalized_predictor, generate_one_denormalized_prediction
            self.generate_one_normalized_predictor = generate_one_normalized_predictor
            self.generate_one_denormalized_prediction = generate_one_denormalized_prediction
            self.trained_model = keras.models.load_model(mld['trained_model'])
            self.dict_normalization_params = pickle.load(open(mld['dict_normalization_params'], 'rb'))
            self.ml_predictors = [self.dict_normalization_params['predictors'][i]['content'] for i in self.dict_normalization_params['predictors']]
        else:
            logger.debug('No machine learning correction available.')

        # Calling general constructor of parent class
        super(OceanDrift, self).__init__(*args, **kwargs)

        self._add_config({
            'drift:vertical_advection': {'type': 'bool', 'default': True, 'description':
                'Advect elements with vertical component of ocean current.',
                'level': self.CONFIG_LEVEL_BASIC},
            'drift:vertical_mixing': {'type': 'bool', 'default': False, 'level': self.CONFIG_LEVEL_BASIC,
                'description': 'Activate vertical mixing scheme with inner loop'},
            'vertical_mixing:timestep': {'type': 'float', 'min': 0.1, 'max': 3600, 'default': 60,
                'level': self.CONFIG_LEVEL_ADVANCED, 'units': 'seconds', 'description':
                'Time step used for inner loop of vertical mixing.'},
            'vertical_mixing:diffusivitymodel': {'type': 'enum', 'default': 'environment',
                'enum': ['environment', 'stepfunction', 'windspeed_Sundby1983',
                 'windspeed_Large1994', 'gls_tke','constant'], 'level': self.CONFIG_LEVEL_ADVANCED,
                 'units': 'seconds', 'description': 'Algorithm/source used for profile of vertical diffusivity. Environment means that diffusivity is aquired from readers or environment constants/fallback.'},
            'vertical_mixing:background_diffusivity': {'type': 'float', 'min': 0, 'max': 1, 'default': 1.2e-5,
                'level': self.CONFIG_LEVEL_ADVANCED, 'units': 'm2s-1', 'description':
                'Background diffusivity used below mixed layer for wind parameterisations.'},
            'vertical_mixing:TSprofiles': {'type': 'bool', 'default': False, 'level':
                self.CONFIG_LEVEL_ADVANCED,
                'description': 'Update T and S profiles within inner loop of vertical mixing. This takes more time, but may be slightly more accurate.'},
            'drift:wind_drift_depth': {'type': 'float', 'default': 0.1,
                'min': 0, 'max': 10, 'units': 'meters',
                'description': 'The direct wind drift (windage) is linearly decreasing from the surface value (wind_drift_factor) until 0 at this depth.',
                'level': self.CONFIG_LEVEL_ADVANCED},
            'drift:stokes_drift': {'type': 'bool', 'default': True,
                'description': 'Advection elements with Stokes drift (wave orbital motion).',
                'level': self.CONFIG_LEVEL_ADVANCED},
            'drift:use_tabularised_stokes_drift': {'type': 'bool', 'default': False,
                'description': 'If True, Stokes drift is estimated from wind based on look-up-tables for given fetch (drift:tabularised_stokes_drift_fetch).',
                'level': self.CONFIG_LEVEL_ADVANCED},
            'drift:tabularised_stokes_drift_fetch': {'type': 'enum', 'enum': ['5000', '25000', '50000'], 'default': '25000',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description':
                'The fetch length when using tabularised Stokes drift.'},
            'general:seafloor_action': {'type': 'enum', 'default': 'lift_to_seafloor',
                'enum': ['none', 'lift_to_seafloor', 'deactivate', 'previous'],
                'description': '"deactivate": elements are deactivated; "lift_to_seafloor": elements are lifted to seafloor level; "previous": elements are moved back to previous position; "none"; seafloor is ignored.',
                'level': self.CONFIG_LEVEL_ADVANCED},
            'drift:truncate_ocean_model_below_m': {'type': 'float', 'default': None,
                'min': 0, 'max': 10000, 'units': 'm',
                'description': 'Ocean model data are only read down to at most this depth, and extrapolated below. May be specified to read less data to improve performance.',
                'level': self.CONFIG_LEVEL_ADVANCED},
             'seed:z': {'type': 'float', 'default': 0,
                    'min': -10000, 'max': 0, 'units': 'm',
                'description': 'Depth below sea level where elements are released. This depth is neglected if seafloor seeding is set selected.',
                'level': self.CONFIG_LEVEL_ESSENTIAL},
            'seed:seafloor': {'type': 'bool', 'default': False,
                'description': 'Elements are seeded at seafloor, and seeding depth (z) is neglected.',
                'level': self.CONFIG_LEVEL_ESSENTIAL},
            })

    def update(self):
        """Update positions and properties of elements."""

        # Simply move particles with ambient current
        self.advect_ocean_current()

        # Advect particles due to surface wind drag,
        # according to element property wind_drift_factor
        self.advect_wind()

        # Stokes drift
        self.stokes_drift()

        # Turbulent Mixing
        if self.get_config('drift:vertical_mixing') is True:
            self.update_terminal_velocity()
            self.vertical_mixing()
        else:  # Buoyancy
            self.vertical_buoyancy()

        # Vertical advection
        self.vertical_advection()

        # Optional machine learning correction
        self.machine_learning_correction()

    def wind_drift_factor_from_trajectory_lw(self, drifters, wind_drift_factors,
                                             simulation_length, simulation_interval):
        """Perform simulations and use skillscore to optimize wind_drift_factor

        drifters: list of disctionaries with numpy arrays of 'lon' and 'lat'
                    and list of datetimes
        wind_drift_factors: the wind_drift_factors to use for simulations/optimalizations
        """

        output = []  # List with one dictionary per trajectory
        for d in drifters:
            time = np.array(d['time'])
            ti = time[1::] - time[0:-1]
            if ti.min() == ti.max():
                trajectory_interval = ti[0]
            else:
                raise ValueError('Trajectory interval is not constant')
            if 'tiv' not in locals():
                tiv = trajectory_interval
                index_interval = simulation_interval.total_seconds()/trajectory_interval.total_seconds()
                if not index_interval.is_integer():
                    raise ValueError('Simulation interval must be a multiple of trajectory time step')
            else:
                if trajectory_interval != tiv:
                    raise ValueError('Trajectories do not have the same time step')
            last_index = (time[-1]-time[0]-simulation_length).total_seconds()/trajectory_interval.total_seconds()
            seed_indices = np.arange(0, last_index, index_interval).astype(int)
            do = {'lon': d['lon'][seed_indices],
                  'lat': d['lat'][seed_indices],
                  'time': np.array(time[seed_indices]),
                  'trajectory_start_index': seed_indices}
            output.append(do)

            if not 'last_seed_time' in locals():
                last_seed_time = do['time'][-1]
            else:
                last_seed_time = np.maximum(do['time'][-1], last_seed_time)
            # Seed for all starting positions in this trajectory
            for lo,la,ti in zip(do['lon'], do['lat'], do['time']):
                ow = np.ones(len(wind_drift_factors))
                self.seed_elements(lon=lo*ow, lat=la*ow, time=ti,
                                   wind_drift_factor=wind_drift_factors)

        self.set_config('drift:max_age_seconds', simulation_length.total_seconds()+1)
        self.run(end_time=last_seed_time+simulation_length)

        index_of_first, index_of_last = \
            self.index_of_activation_and_deactivation()
        times_model = np.array(self.get_time_array()[0])

        #import matplotlib.pyplot as plt
        i = 0  # element number in simulation
        for d, o in zip(drifters, output):
            o['segments'] = {}
            o['wind_drift_factor'] = np.ones(len(o['trajectory_start_index']))
            o['skillscore'] = np.ones(len(o['trajectory_start_index']))
            for segnum, tsi in enumerate(o['trajectory_start_index']):
                o['segments'][segnum] = {'skillscore': ow}
                lon_obs = d['lon'][tsi:tsi+int(index_interval)]
                lat_obs = d['lat'][tsi:tsi+int(index_interval)]
                #plt.plot(lon_obs, lat_obs, 'r')
                for wnum, wdf in enumerate(wind_drift_factors):
                    lon_model = self.history['lon'][i, index_of_first[i]:index_of_last[i]]
                    lat_model = self.history['lat'][i, index_of_first[i]:index_of_last[i]]
                    i = i + 1
                    ss = skillscore_liu_weissberg(lon_obs, lat_obs, lon_model, lat_model)
                    o['segments'][segnum]['skillscore'][wnum] = ss
                    #plt.plot(lon_model, lat_model, label='Skillscore: %s' % ss)

                print(o['segments'][segnum]['skillscore'], 'SS')
                o['wind_drift_factor'][segnum] = \
                    wind_drift_factors[np.argmax(o['segments'][segnum]['skillscore'])]
                o['skillscore'][segnum] = o['segments'][segnum]['skillscore'].max()
                #plt.legend()
                #plt.show()
                #stop

        #import xarray as xr
        #da = xr.DataArray(
        #    data=skillscore,
        #    dims=["wind_drift_factor", "time"],
        #    coords=dict(
        #        wind_drift_factor=wind_drift_factors,
        #        time=time
        #    ),
        #    attrs=dict(
        #        description="Liu Weissberg skillscore",
        #    ),
        #)

        return output

    def machine_learning_correction(self):
        if not hasattr(self, 'trained_model'):
            return  # No machine learning correction available

        # ML shall only be applied every 1 hour
        if not hasattr(self, 'ml_timesteps'):
            self.ml_timesteps = 3600/self.time_step.total_seconds()
        if self.steps_calculation % self.ml_timesteps != 0:
            logger.warning('ML not applied at time step %s' % self.steps_calculation)
            return

        # Only apply every 1 hour
        logger.warning('Machine learning')
        x_vel = np.ones(self.num_elements_active())*np.NaN
        y_vel = np.ones(self.num_elements_active())*np.NaN
        st = datetime.now()

        list_predictors_normalized = []
        list_predictors_denormalized = []

        # Prepare predictor dicts
        for n in range(self.num_elements_active()):
            dict_input = {s: getattr(self.environment, s)[n] for s in self.ml_predictors}
            predictor_normalized = self.generate_one_normalized_predictor(
                                            dict_input, self.dict_normalization_params)
            list_predictors_normalized.append(predictor_normalized)

        # Normalize
        all_predictors_normalized = np.squeeze(np.array(list_predictors_normalized))
        # perform a prediction from normalized predictor to normalized label
        logger.warning('Calculating ML correction...')
        predicted_normalized_residual_correction = self.trained_model.predict(all_predictors_normalized)
        # get the "native units" residual correction
        logger.warning('Denormalization...')
        for n in range(self.num_elements_active()):
            native_units_correction = self.generate_one_denormalized_prediction(
                predicted_normalized_residual_correction[n, :], self.dict_normalization_params)
            x_vel[n] = native_units_correction['residual_displacement_x']*1000/3600
            y_vel[n] = native_units_correction['residual_displacement_y']*1000/3600

        # Apply correction
        logger.warning('Applying ML correction: %s to %s m/s eastwards, %s to %s m/s northwards' %
                        (x_vel.min(), x_vel.max(), y_vel.min(), y_vel.max()))
        #logger.warning('%s particles, %s' % (self.num_elements_active(), datetime.now()-st))
        self.update_positions(x_vel, y_vel)

    def disable_vertical_motion(self):
        """Deactivate any vertical processes/advection"""
        conf = {
                'drift:vertical_advection': False,
                'drift:vertical_mixing': False}
        for co, va in conf.items():
            logger.info('Setting config: %s -> %s' % (co, va))
            self.set_config(co, va)

    def update_terminal_velocity(self, Tprofiles=None, Sprofiles=None,
                                 z_index=None):
        """Calculate terminal velocity due to bouyancy from own properties
        and environmental variables. Sub-modules should overload
        this method for particle-specific behaviour
        """
        pass

    def prepare_vertical_mixing(self):
        pass  # To be implemented by subclasses as needed

    def vertical_advection(self):
        """Move particles vertically according to vertical ocean current

            Vertical advection by ocean currents is normally small
            compared to termical velocity
        """
        if self.get_config('drift:vertical_advection') is False:
            logger.debug('Vertical advection deactivated')
            return

        in_ocean = np.where(self.elements.z<0)[0]
        if len(in_ocean) > 0:
            w = self.environment.upward_sea_water_velocity[in_ocean]
            self.elements.z[in_ocean] = np.minimum(0,
                self.elements.z[in_ocean] + self.elements.moving[in_ocean] * w * self.time_step.total_seconds())
        else:
            logger.debug('No vertical advection for elements at surface')

    def vertical_buoyancy(self):
        """Move particles vertically according to their buoyancy"""
        in_ocean = np.where(self.elements.z<0)[0]
        if len(in_ocean) > 0:
            self.elements.z[in_ocean] = np.minimum(0,
                self.elements.z[in_ocean] + self.elements.terminal_velocity[in_ocean] * self.time_step.total_seconds())

        # check for minimum height/maximum depth for each particle
        Zmin = -1.*self.environment.sea_floor_depth_below_sea_level

        # Let particles stick to bottom
        bottom = np.where(self.elements.z < Zmin)
        if len(bottom[0]) > 0:
            logger.debug('%s elements reached seafloor, set to bottom' % len(bottom[0]))
            self.interact_with_seafloor()
            self.bottom_interaction(Zmin)

    def surface_stick(self):
        '''To be overloaded by subclasses, e.g. downward mixing of oil'''

        # keep particle just below the surface
        surface = np.where(self.elements.z >= 0)
        if len(surface[0]) > 0:
            self.elements.z[surface] = -0.01

    def bottom_interaction(self, Zmin=None):
        '''To be overloaded by subclasses, e.g. radionuclides in sediments'''
        pass

    def surface_wave_mixing(self, time_step_seconds):
        '''To be overloaded by subclasses, e.g. downward mixing of oil'''

        pass

    def get_diffusivity_profile(self, model):
        depths = np.abs(self.environment_profiles['z'])
        wind, depth = np.meshgrid(self.wind_speed(), depths)
        MLD = self.environment.ocean_mixed_layer_thickness
        background_diffusivity = self.get_config('vertical_mixing:background_diffusivity')

        if model == 'windspeed_Large1994':
            return verticaldiffusivity_Large1994(wind, depth, MLD, background_diffusivity)
        elif model == 'windspeed_Sundby1983':
            return verticaldiffusivity_Sundby1983(wind, depth, MLD, background_diffusivity)
        elif model == 'gls_tke':
            if not hasattr(self, 'gls_parameters'):
                logger.info('Searching readers for GLS parameters...')
                for reader_name, reader in self.readers.items():
                    if hasattr(reader, 'gls_parameters'):
                        self.gls_parameters = reader.gls_parameters
                        logger.info('Found gls-parameters in ' + reader_name)
                        break  # Success
                if not hasattr(self, 'gls_parameters'):
                    logger.info('Did not find gls-parameters in any readers.')
                    self.gls_parameters = None
            windstress = np.sqrt(self.environment.surface_downward_x_stress**2 +
                                 self.environment.surface_downward_y_stress**2)
            return gls_tke(windstress, depth, self.sea_water_density(),
                           self.environment.turbulent_kinetic_energy,
                           self.environment.turbulent_generic_length_scale,
                           gls_parameters)

        else:
            raise ValueError('Unknown diffusivity model: ' + model)

    def vertical_mixing(self, store_depths=False):
        """Mix particles vertically according to eddy diffusivity and buoyancy

            Buoyancy is expressed as terminal velocity, which is the
            steady-state vertical velocity due to positive or negative
            buoyant behaviour. It is usually a function of particle density,
            diameter, and shape.

            Vertical particle displacemend du to turbulent mixing is
            calculated using a random walk scheme" (Visser et al. 1996)
        """

        if self.get_config('drift:vertical_mixing') is False:
            logger.debug('Turbulent mixing deactivated')
            return

        self.timer_start('main loop:updating elements:vertical mixing')

        dt_mix = self.get_config('vertical_mixing:timestep')

        # minimum height/maximum depth for each particle
        Zmin = -1.*self.environment.sea_floor_depth_below_sea_level

        # Eventual model specific preparions
        self.prepare_vertical_mixing()

        # get profile of eddy diffusivity
        # get vertical eddy diffusivity from environment or specific model
        diffusivity_model = self.get_config('vertical_mixing:diffusivitymodel')
        if diffusivity_model == 'environment':
            if not hasattr(self, 'fallback_values'):
                self.set_fallback_values()
            if 'ocean_vertical_diffusivity' in self.environment_profiles and not (self.environment_profiles['ocean_vertical_diffusivity'].min() == self.fallback_values['ocean_vertical_diffusivity'] and self.environment_profiles['ocean_vertical_diffusivity'].max() == self.fallback_values['ocean_vertical_diffusivity']):
                Kprofiles = self.environment_profiles[
                    'ocean_vertical_diffusivity']
                logger.debug('Using diffusivity from ocean model')
            else:
                logger.debug('Using diffusivity from Large1994 since model diffusivities not available')
                # Using higher vertical resolution when analytical
                self.environment_profiles['z'] = \
                        -np.arange(0, self.environment.ocean_mixed_layer_thickness.max() + 2)
                Kprofiles = self.get_diffusivity_profile('windspeed_Large1994')
        elif diffusivity_model == 'constant':
            logger.debug('Using constant diffusivity specified by fallback_values[''ocean_vertical_diffusivity''] = %s m2.s-1' % (self.fallback_values['ocean_vertical_diffusivity']))
            Kprofiles = self.fallback_values['ocean_vertical_diffusivity']*np.ones(
                    self.environment_profiles['ocean_vertical_diffusivity'].shape) # keep constant value for ocean_vertical_diffusivity
        else:
            logger.debug('Using functional expression for diffusivity')
            # Using higher vertical resolution when analytical
            if self.environment_profiles is None:
                self.environment_profiles = {}
            self.environment_profiles['z'] = \
                    -np.arange(0, self.environment.ocean_mixed_layer_thickness.max() + 2)
            # Note: although analytical functions, z is discretised
            Kprofiles = self.get_diffusivity_profile(diffusivity_model)

        logger.debug('Diffusivities are in range %s to %s' %
                      (Kprofiles.min(), Kprofiles.max()))

        # get profiles of salinity and temperature
        # (to save interpolation time in the inner loop)
        if (self.get_config('vertical_mixing:TSprofiles') is True
            and 'sea_water_salinity' in self.required_variables):
            Sprofiles = self.environment_profiles['sea_water_salinity']
            Tprofiles = \
                self.environment_profiles['sea_water_temperature']
            if ('sea_water_salinity' in self.fallback_values and
                Sprofiles.min() == Sprofiles.max()):
                logger.debug('Salinity and temperature are fallback'
                              'values, skipping TSprofile')
                Sprofiles = None
                Tprofiles = None
            else:
                logger.debug('Using TSprofiles for vertical mixing')
        else:
            logger.debug('TSprofiles deactivated for vertical mixing')
            Sprofiles = None
            Tprofiles = None

        # prepare vertical interpolation coordinates
        z_i = range(self.environment_profiles['z'].shape[0])
        if len(z_i) == 1:
            z_index = 0
        else:
            z_index = interp1d(-self.environment_profiles['z'],
                               z_i, bounds_error=False,
                               fill_value=(0,len(z_i)-1))  # Extrapolation

        # Internal loop for fast time step of vertical mixing model.
        # Random walk needs faster time step than horizontal advection.
        logger.debug('Vertical mixing module:' +
            self.get_config('vertical_mixing:diffusivitymodel'))
        ntimes_mix = np.abs(int(self.time_step.total_seconds()/dt_mix))
        logger.debug('Turbulent diffusion with random walk '
                      'scheme using ' + str(ntimes_mix) +
                      ' fast time steps of dt=' + str(dt_mix) + 's')

        if store_depths is not False:
            depths = np.zeros((ntimes_mix, self.num_elements_active()))
            depths[0, :] = self.elements.z

        # Calculating dK/dz for all profiles before the loop
        gradK = -np.gradient(Kprofiles, self.environment_profiles['z'], axis=0)
        gradK[np.abs(gradK)<1e-10] = 0

        for i in range(0, ntimes_mix):
            #remember which particles belong to the exact surface
            surface = self.elements.z == 0

            # Update the terminal velocity of particles
            self.update_terminal_velocity(Tprofiles=Tprofiles, Sprofiles=Sprofiles, z_index=z_index)
            w = self.elements.terminal_velocity

            # Diffusivity and its gradient at z
            zi = np.round(z_index(-self.elements.z)).astype(np.uint16)
            Kz = Kprofiles[zi, range(Kprofiles.shape[1])]
            dKdz = gradK[zi, range(Kprofiles.shape[1])]

            # Visser et al. 1997 random walk mixing
            # requires an inner loop time step dt such that
            # dt << (d2K/dz2)^-1, e.g. typically dt << 15min
            #
            # NB: In the last term Kz is evaluated in zi, while
            # it should be evaluated in (self.elements.z - dKdz*dt_mix)
            # This is not expected have large impact on the result
            R = 2*np.random.random(self.num_elements_active()) - 1
            r = 1.0/3
            # New position  =  old position   - up_K_flux   + random walk
            self.elements.z = self.elements.z - self.elements.moving*(
                dKdz*dt_mix - R*np.sqrt((Kz*dt_mix*2/r)))

            # Reflect from surface
            reflect = np.where(self.elements.z >= 0)
            if len(reflect[0]) > 0:
                self.elements.z[reflect] = -self.elements.z[reflect]

            # Reflect elements going below seafloor
            bottom = np.where(np.logical_and(self.elements.z < Zmin, self.elements.moving == 1))
            if len(bottom[0]) > 0:
                logger.debug('%s elements penetrated seafloor, lifting up' % len(bottom[0]))
                self.elements.z[bottom] = 2*Zmin[bottom] - self.elements.z[bottom]

            # Advect due to buoyancy
            self.elements.z = self.elements.z + w*dt_mix*self.elements.moving

            # Put the particles that belonged to the surface slick
            # (if present) back to the surface
            self.elements.z[surface] = 0.

            # Formation of slick and wave mixing for surfaced particles
            # if implemented for this class
            self.surface_stick()
            self.surface_wave_mixing(dt_mix)

            # Let particles stick to bottom
            bottom = np.where(self.elements.z < Zmin)
            if len(bottom[0]) > 0:
                logger.debug('%s elements reached seafloor, set to bottom' % len(bottom[0]))
                self.interact_with_seafloor()
                self.bottom_interaction(Zmin)

            if store_depths is not False:
                depths[i, :] = self.elements.z

        self.timer_end('main loop:updating elements:vertical mixing')

        if store_depths is not False:
            return depths
        else:
            return None

    def animate_vertical_distribution(self, depths=None, maxdepth=None, bins=50, filename=None, subsamplingstep=1):
        """Function to animate vertical distribution of particles
            bins:            number of bins in the histogram
            maxdepth:        maximum depth
            subsamplingstep: speed-up the generation of the animation reducing the number of output frames
            fasterwriter:    speed-up the writing to outpute file
        """
        import matplotlib.pyplot as plt
        import matplotlib.animation as animation

        start_time = datetime.now()

        #from timeit import default_timer as timer
        #start=timer()

        fig, (axk, axn) = plt.subplots(1, 2, gridspec_kw={'width_ratios': [1, 3]})
        if depths is not None:  # Debug mode, output from one cycle has been provided
            z = depths
            time_step = self.get_config('vertical_mixing:timestep')
            times = [self.time + i*timedelta(seconds=time_step) for i in range(z.shape[0])]
        else:
            z = self.get_property('z')[0]
            z = np.ma.filled(z, np.nan)
            K = self.get_property('ocean_vertical_diffusivity')[0]
            time_step = self.time_step.total_seconds()
            times = self.get_time_array()[0]
        if maxdepth is None:
            maxdepth = np.nanmin(z)
        if maxdepth > 0:
            maxdepth = -maxdepth  # negative z

        if depths is not None:
            axk.plot(np.nanmean(self.environment_profiles['ocean_vertical_diffusivity'], 1), self.environment_profiles['z'])
            xmax = self.environment_profiles['ocean_vertical_diffusivity'].max()
        else:
            axk.plot(K, z, 'k.')
            xmax = np.nanmax(K)
        axk.set_ylim([maxdepth, 0])
        axk.set_xlim([0, xmax*1.1])
        axk.set_ylabel('Depth [m]')
        axk.set_xlabel('Vertical diffusivity [$m^2/s$]')

        subsamplingstep=max(1,subsamplingstep)
        times=times[0:-1:subsamplingstep]
        z=z[0:-1:subsamplingstep,:]

        hist_series = np.zeros((bins, len(times)))
        bin_series = np.zeros((bins+1, len(times)))
        for i in range(len(times)):
            hist_series[:,i], bin_series[:,i] = np.histogram(z[i,:][np.isfinite(z[i,:])], bins=bins, range=(maxdepth,z[np.isfinite(z)].max()))
        maxnum = hist_series.max()

        axn.clear()
        bc=axn.barh(bin_series[0:-1,i], hist_series[:,i], height=-maxdepth/bins, align='edge')
        axn.set_ylim([maxdepth, 0])
        axn.set_xlim([0, maxnum])
        title=axn.set_title('%s UTC' % times[i])
        axn.set_xlabel('Number of particles')

        def update_histogram(i):
            for rect, y in zip(bc, hist_series[:,i]):
                rect.set_width(y)
            title.set_text('%s UTC' % times[i])

        self.__save_or_plot_animation__(plt.gcf(),
                                        update_histogram,
                                        filename,
                                        frames = len(times),
                                        fps = 10,
                                        interval=50,
                                        blit=False)

        logger.info('Time to make animation: %s' %
                    (datetime.now() - start_time))




    def plot_vertical_distribution(self, maxdepth=None, bins=None, maxnum=None):
        """Function to plot vertical distribution of particles

            maxdepth: maximum depth considered for the profile
            bins:     number of bins between surface and maxdepth
            maxnum:   range of bars in histogram is [0,maxnum]
        """
        import matplotlib.pyplot as plt
        from matplotlib.widgets import Slider

        fig = plt.figure()
        mainplot = fig.add_axes([.15, .3, .8, .5])
        sliderax = fig.add_axes([.15, .08, .75, .05])
        tslider = Slider(sliderax, 'Timestep', 0, self.steps_output-1,
                         valinit=self.steps_output-1, valfmt='%0.0f')
        try:
            dz = self.get_config('vertical_mixing:verticalresolution')
        except:
            dz = 1.
        maxrange = -100

        # overwrite default values if input arguments are provided
        if maxdepth is not None:
            maxrange = maxdepth
        if maxrange > 0:
            maxrange = -maxrange  # negative z

        if bins is not None:
            dz = -maxrange/bins

        # using get_property instead of history to exclude elements thare are not yet seeded
        z = self.get_property('z')[0]
        z = np.ma.filled(z, np.nan)

        if maxnum is None:
            # Precalculatig histograms to find maxnum
            hist_series = np.zeros((int(-maxrange/dz), self.steps_output-1))
            bin_series = np.zeros((int(-maxrange/dz)+1, self.steps_output-1))
            for i in range(self.steps_output-1):
                hist_series[:,i], bin_series[:,i] = np.histogram(z[i,:][np.isfinite(z[i,:])], bins=int(-maxrange/dz), range=[maxrange, 0])
            maxnum = hist_series.max()

        def update(val):
            tindex = int(tslider.val)
            mainplot.cla()
            mainplot.grid()
            mainplot.hist(z[tindex, :], bins=int(-maxrange/dz),
                          range=[maxrange, 0], orientation='horizontal')
            mainplot.set_ylim([maxrange, 0])
            mainplot.set_xlim([0, maxnum])
            mainplot.set_xlabel('number of particles')
            mainplot.set_ylabel('depth [m]')
            x_wind = self.history['x_wind'].T[tindex, :]
            y_wind = self.history['y_wind'].T[tindex, :]
            windspeed = np.mean(np.sqrt(x_wind**2 + y_wind**2))
            mainplot.set_title(str(self.get_time_array()[0][tindex]) +
                               #'   Percent at surface: %.1f %' % percent_at_surface)
                               '   Mean windspeed: %.1f m/s' % windspeed)
            fig.canvas.draw_idle()

        update(0)  # Plot initial distribution
        tslider.on_changed(update)
        plt.show()

        # returning objects prevents unresponsiveness when moving the Slider
        # https://github.com/matplotlib/matplotlib/issues/3105#issuecomment-44855888
        return fig,mainplot,sliderax,tslider

    def plotter_vertical_distribution_time(self, ax=None, mask=None,
            dz=1., maxrange=-100, bins=None, step=1):
        """Function to plot vertical distribution of particles.

        Use mask to plot any selection of particles.
        """
        from pylab import axes, draw
        from matplotlib import dates, pyplot

        if ax is None:
            fig = pyplot.figure()
            ax = fig.gca()
            show = True
        else:
            show = False

        if mask is None: # create a mask that is True for all particles
            mask = self.history['z'].T[0] == self.history['z'].T[0]

        if bins is None:
            bins=int(-maxrange/dz)

        ax.hist(self.history['z'].T[step,mask], bins=bins,
                range=[maxrange, 0], orientation='horizontal')
        ax.set_ylim([maxrange, 0])
        ax.grid()
        #ax.set_xlim([0, mask.sum()*.15])
        ax.set_xlabel('Number of particles')
        ax.set_ylabel('Depth [m]')
        x_wind = self.history['x_wind'].T[step, :]
        y_wind = self.history['x_wind'].T[step, :]
        windspeed = np.mean(np.sqrt(x_wind**2 + y_wind**2))
        ax.set_title(str(self.get_time_array()[0][step]) +
                     '   Mean windspeed: %.1f m/s' % windspeed)
        if show is True:
            pyplot.show()



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2016, Knut-Frode Dagestad, MET Norway

import logging; logger = logging.getLogger(__name__)
from datetime import timedelta
import numpy as np
from math import sqrt
import matplotlib.pyplot as plt
import pyproj
import cmocean


def wind_drift_factor_from_trajectory(trajectory_dict, min_period=None):
    '''Estimate wind_drift_fator based on wind and current along given trajectory

    trajectory_dict: dictionary with arrays of same length of the following variables:
    time, lon, lat, x_wind, y_wind, x_sea_water_velocity, y_sea_water_velocity

    Returns array of same length minus one of the fitted wind_drift_factor
    '''

    geod = pyproj.Geod(ellps='WGS84')
    time = trajectory_dict['time']
    try:
        import pandas as pd
        time = pd.to_datetime(time)
    except:
        pass
    if min_period is None:
        ind = range(len(time))
    else:
        timestep = time[1] - time[0]
        s = np.round(min_period.total_seconds()/(timestep).total_seconds()).astype(int)
        ind = np.arange(0, len(time), s).astype(np.int)
        print('Original timestep (%s) multiplied by %i: %s' % (timestep, s, timestep*s))
        ind2 = ind.copy()
        for i in range(1, s):
            ind2 = np.concatenate((ind2, ind+i))
        ind = ind2
        ind = ind[ind<len(time)]
        time = time[ind]
    cu = trajectory_dict['x_sea_water_velocity'][ind]
    cv = trajectory_dict['y_sea_water_velocity'][ind]
    wu = trajectory_dict['x_wind'][ind]
    wv = trajectory_dict['y_wind'][ind]
    lon = trajectory_dict['lon'][ind]
    lat = trajectory_dict['lat'][ind]
    current_speed = np.sqrt(cu**2+cv**2)
    current_azimuth = np.degrees(np.arctan2(cu, cv))
    wind_speed = np.sqrt(wu**2+wv**2)
    time_step = (time[1]-time[0]).total_seconds()
    # Move forward from each position with current
    lonf, latf, back_az = geod.fwd(lon[0:-1], lat[0:-1], current_azimuth[0:-1], current_speed[0:-1]*time_step)
    # Find distance and azimuth to next observational position
    azimuth_forward, a2, distance = geod.inv(lonf, latf, lon[1:], lat[1:])
    # Find the wind_drift_factor needed to give the remaining distance
    wind_drift_factor = distance / (time_step*wind_speed[0:-1])

    wind_azimuth = np.degrees(np.arctan2(wu, wv))
    azimuth_offset = azimuth_forward - wind_azimuth[0:-1]  # 0 if downwind drift, positive if rightwards drift is needed towards end position
    azimuth_offset = azimuth_offset % 360  # Make sure azimuth angle is between -180 and 180
    azimuth_offset = (azimuth_offset + 360) % 360
    azimuth_offset[azimuth_offset>180] -= 360

    return wind_drift_factor, azimuth_offset

def distance_between_trajectories(lon1, lat1, lon2, lat2):
    '''Calculate the distances [m] between two trajectories'''

    assert len(lon1) == len(lat1) == len(lat1) == len(lat2)
    geod = pyproj.Geod(ellps='WGS84')
    azimuth_forward, a2, distance = geod.inv(lon1, lat1, lon2, lat2)
 
    return distance

def distance_along_trajectory(lon, lat):
    '''Calculate the distances [m] between points along a trajectory'''

    geod = pyproj.Geod(ellps='WGS84')
    azimuth_forward, a2, distance = geod.inv(lon[1:], lat[1:], lon[0:-1], lat[0:-1])

    return distance

def skillscore_darpa(lon1, lat1, lon2, lat2):
    '''Scoring algorithm used for DARPA float challenge 2021.

    Copied from implementation made by Jean Rabault.
    Assuming 6 positions, separated by 2 days, where first pos (start) are identical.
    '''

    assert len(lon1) == 6
    distances = distance_between_trajectories(lon1, lat1, lon2, lat2)/1000
    if distances[0] != 0:
        logger.warning('DARPA score: first position is not identical '
                       '(distance: %s km)' % distances[0])
    distances = distances[1:]  # For the remaining, we ignore starting position

    dict_DARPA_points = {
        4: 5,
        8: 2,
        16: 1,
        32: 0.25 }
    distance_thresholds = sorted(list(dict_DARPA_points.keys()))

    dict_DARPA_scoring_multiplicator = {
        0: 1,
        1: 2,
        2: 5,
        3: 10,
        4: 20 }

    DARPA_points = 0
    for crrt_ind, crrt_distance in enumerate(distances):
        if crrt_distance >= distance_thresholds[-1]:
            break

        for crrt_threshold in distance_thresholds:
            if crrt_distance < crrt_threshold:
                crrt_points = dict_DARPA_points[crrt_threshold]
                break

        crrt_multiplicator = dict_DARPA_scoring_multiplicator[crrt_ind]

        DARPA_points += int(crrt_points * crrt_multiplicator)

    return DARPA_points

def skillscore_liu_weissberg(lon_obs, lat_obs, lon_model, lat_model, tolerance_threshold=1):
    ''' calculate skill score from normalized cumulative seperation
    distance. Liu and Weisberg 2011.

    Returns the skill score bewteen 0. and 1.
    '''

    lon_obs = np.array(lon_obs)
    lat_obs = np.array(lat_obs)
    lon_model = np.array(lon_model)
    lat_model = np.array(lat_model)
    d = distance_between_trajectories(lon_obs, lat_obs, lon_model, lat_model)
    l = distance_along_trajectory(lon_obs, lat_obs)
    s = d.sum() / l.cumsum().sum()
    if tolerance_threshold==0:
        skillscore = 0
    else:
        skillscore = max(0, 1 - s/tolerance_threshold)

    return skillscore

def plot_wind_drift_factor(wdf, azimuth, wmax_plot=None):
    '''Polar plot of array of wind drift factor, with associated azimuthal offset'''

    wmax = wdf.max()
    wbins = np.arange(0, wmax+.005, .005)
    abins = np.linspace(-180, 180, 30)
    hist, _, _ = np.histogram2d(azimuth, wdf, bins=(abins, wbins))
    A, W = np.meshgrid(abins, wbins)
    fig, ax = plt.subplots(subplot_kw=dict(projection='polar'))
    ax.set_theta_zero_location('N', offset=0)
    ax.set_theta_direction(-1)
    pc = ax.pcolormesh(np.radians(A), W, hist.T, cmap=cmocean.cm.dense)
    plt.arrow(np.pi, wmax, 0, -wmax, width=.015, facecolor='k', zorder=100,
              head_width=.8, lw=2, head_length=.005, length_includes_head=True)
    plt.text(np.pi, wmax*.5, ' Wind direction', fontsize=18)
    plt.grid()
    if wmax_plot is not None:
        plt.ylim([0, wmax_plot])
    plt.show()


def oil_wave_entrainment_rate_li2017(dynamic_viscosity, oil_density, interfacial_tension,
                                     significant_wave_height=None, wave_breaking_fraction=None,
                                     wind_speed=None, sea_water_density=1028.):
    # Z. Li, M.L. Spaulding, D. French McCay, J. Mar. Pollut. Bull. (2016):
    # An algorithm for modeling entrainment and naturally and chemically dispersed
    # oil droplet size distribution under surface breaking wave conditions

    if wave_breaking_fraction is None:
        if wind_speed is None:
            raise ValueError('wave_breaking_fraction or wind_speed must be provided')
        wave_breaking_fraction = wave_breaking_fraction_from_wind(wind_speed)
    if significant_wave_height is None:
        if wind_speed is None:
            raise ValueError('significant_wave_height or wind_speed must be provided')
        significant_wave_height = significant_wave_height_from_wind_neumann_pierson(wind_speed)
    g = 9.81
    delta_rho = sea_water_density - oil_density
    d_o = 4*np.sqrt(interfacial_tension / (delta_rho*g))
    we = sea_water_density*g*significant_wave_height*d_o/interfacial_tension
    oh = dynamic_viscosity/np.sqrt(oil_density*interfacial_tension*d_o)
    with np.errstate(divide='ignore'):
        entrainment_rate = (4.604e-10*we**1.805*oh**-1.023)*wave_breaking_fraction
    return entrainment_rate

def significant_wave_height_from_wind_neumann_pierson(wind_speed):
    # Neumann and Pierson, 1966
    # WMO 1998
    return 0.0246*np.power(wind_speed, 2)

def wave_breaking_fraction_from_wind(wind_speed, wave_period=None):
    # TODO: We should also have an option here for
    # the case when wave height is given, but no wind
    if wave_period is None:
        wave_period = wave_period_from_wind(wind_speed)
    f = 0.032*(wind_speed - 5)/wave_period
    f[f < 0] = 0
    return f

def wave_period_from_wind(wind_speed):
    # Pierson-Moskowitz if period not available from readers
    # WMO guide to wave analysis and forecasting pp. 14, WMO (1998)
    # fallback value if no wind speed or Hs to avoid division by zero
    wind_speed = np.atleast_1d(wind_speed)
    omega = 5*np.ones(wind_speed.shape)  # Angular frequency
    omega[wind_speed>0] = 0.877*9.81/(1.17*wind_speed[wind_speed>0])
    return 2*np.pi/omega

def verticaldiffusivity_Sundby1983(windspeed, depth, mixedlayerdepth=50, background_diffusivity=0):
    ''' Vertical diffusivity from Sundby (1983)

    S. Sundby (1983): A one-dimensional model for the vertical
        distribution of pelagic fish eggs in the mixed layer
        Deep Sea Research (30) pp. 645-661
    '''

    K = 76.1e-4 + 2.26e-4 * windspeed*windspeed * np.ones(np.atleast_1d(depth.shape))
    K[depth>mixedlayerdepth-1] = (K[depth>mixedlayerdepth-1]+background_diffusivity)/2  # Transition
    K[depth>=mixedlayerdepth] = background_diffusivity  # Cutoff below mixed layer
    # valid = windspeed_squared < 13**2
    return K

def verticaldiffusivity_Large1994(windspeed, depth, mixedlayerdepth=50, background_diffusivity=0):
    ''' Vertical diffusivity from Large et al. (1994)

    Depending on windspeed, depth and mixed layer depth (default 50m).'''

    # Defining two helper methods:

    def stabilityfunction(sigma):
        # controls stratification regimes for diffusivity
        # a value of 1 represent neutrally buoyant conditions (no stratification)
        # a value between 0 and 1 represents stable stratification
        return 0.2 # should be depth dependent, too

    def G(sigma):
        # vertical shape function for eddy diffusivity
        a1 = 1.
        a2 = -2
        a3 = 1
        G = a1*sigma + a2*sigma**2 + a3*sigma**3
        below = np.where(G>=1)
        G[below]=G[below]*0.
        return G

    depth = np.abs(depth)  # Making sure depth is positive value
    MLD = mixedlayerdepth  # shorthand
    rhoa = 1.22  # Air density
    cd = 1.25e-3  # Kara et al. 2007
    windstress = windspeed*windspeed * cd * rhoa

    K = MLD * stabilityfunction(depth/MLD) * 0.4 * G(depth/MLD) * windstress
    K[depth>=MLD] = background_diffusivity

    return K

def verticaldiffusivity_stepfunction(depth, MLD=20,
                                     k_above=.1, k_below=.02):
    ''' eddy diffusivity with discontinuity for testing of mixing scheme'''
    depth = np.abs(depth)  # Making sure depth is positive value
    K = k_above*np.ones(depth.shape)
    K[depth>MLD] = k_below
    return K

def gls_tke(windstress, depth, sea_water_density,
            tke, generic_length_scale, gls_parameters=None):
    '''From LADIM model.'''

    g = 9.81
    f0 = 0.1  # mean wave frequency
    c_w = 4.0  # wave mixing parameter
    c_i = 0.2  # coefficient for the interior
    if gls_parameters is None:
        # GLS parameters from ROMS, k-omega closure (see ocean.in)
        p = 0.0
        m = 1.0
        n = 1.0
        cmu0 = 0.5477  # for KANTHA_CLAYSON stability function
    else:
        p = gls_parameters['gls_p']
        m = gls_parameters['gls_m']
        n = gls_parameters['gls_n']
        cmu0 = gls_parameters['gls_cmu0']

    phi = 100. * (windstress/sea_water_density)**(3./2.)

    # dissipation and turbulent length scale for interiour of mixed layer
    eps = cmu0**(3.+p/n)*tke**(3./2.+m/n)*generic_length_scale**(-1./n)
    l_i = c_i * tke**(3./2.) * eps**(-1.)

    # diffusivity for interior of mixed layer
    # c_i = sqrt(2.) * cmu0**3
    ki = c_i * (2.*tke)**0.5 * l_i

    # length scale and diffusivity of wave-enhanced layer
    l_w = np.sqrt(phi / (g*f0))
    kwave = c_w * (2*tke)**0.5 * l_w
    kmix = ki + kwave

    K, N = np.meshgrid(kmix, depths)

    return K


def stokes_drift_profile_breivik(stokes_u_surface, stokes_v_surface,
                                 significant_wave_height, mean_wave_period, z):
    """
    Calculate vertical Stokes drift profile from
    Breivik et al. 2016, A Stokes drift approximation
    based on the Phillips spectrum, Ocean Mod. 100
    """
    stokes_surface_speed = np.sqrt(stokes_u_surface**2 +
                                   stokes_v_surface**2)

    fm02 = fm02 = 1. / mean_wave_period

    total_transport = (2.*np.pi/16.)*fm02*np.power(
                       significant_wave_height, 2)

    k = (stokes_surface_speed/(2*total_transport))

    stokes_speed = stokes_surface_speed*np.exp(2*k*z)

    zeromask = stokes_surface_speed == 0
    stokes_u = stokes_speed*stokes_u_surface/stokes_surface_speed
    stokes_v = stokes_speed*stokes_v_surface/stokes_surface_speed
    stokes_u[zeromask] = 0
    stokes_v[zeromask] = 0

    return stokes_u, stokes_v, stokes_speed


def ftle(X, Y, delta, duration):
    """Calculate Finite Time Lyapunov Exponents"""
    # From Johannes Rohrs
    nx = X.shape[0]
    ny = X.shape[1]
    J = np.empty([nx,ny,2,2], np.float32)
    FTLE = np.empty([nx,ny], np.float32)

    # gradient
    dx = np.gradient(X)
    dy = np.gradient(Y)

    # Jacobian
    J[:,:,0,0] = dx[0] / (2*delta)
    J[:,:,1,0] = dy[0] / (2*delta)
    J[:,:,0,1] = dx[1] / (2*delta)
    J[:,:,1,1] = dy[1] / (2*delta)

    for i in range(0,nx):
        for j in range(0,ny):
            # Green-Cauchy tensor
            D = np.dot(np.transpose(J[i,j]), J[i,j])
            # its largest eigenvalue
            lamda = np.linalg.eigvals(D)
            FTLE[i,j] = np.log(np.sqrt(max(lamda)))/np.abs(duration)

    return FTLE

class PhysicsMethods:
    """Physics methods to be inherited by OpenDriftSimulation class"""

    @staticmethod
    def sea_water_density(T=10., S=35.):
        '''The function gives the density of seawater at one atmosphere
        pressure as given in :

        N.P. Fofonoff and R.C. Millard Jr.,1983,
        Unesco technical papers in marine science no. 44.

        S   = Salinity in promille of the seawater
        T   = Temperature of the seawater in degrees Celsius
        '''

        R4 = 4.8314E-04
        DR350 = 28.106331

        # Pure water density at atmospheric pressure
        # Bigg P.H. (1967) BR. J. Applied Physics pp.:521-537

        R1 = ((((6.536332E-09 * T - 1.120083E-06) * T + 1.001685E-04) *
              T - 9.095290E-03) * T + 6.793952E-02) * T - 28.263737

        # Seawater density at atmospheric pressure
        # coefficients involving salinity :

        R2 = (((5.3875E-09 * T - 8.2467E-07) * T + 7.6438E-05) *
              T - 4.0899E-03) * T + 8.24493E-01

        R3 = (-1.6546E-06*T+1.0227E-04)*T-5.72466E-03

        # International one-atmosphere equation of state of seawater :

        SIG = R1 + (R4*S + R3*np.sqrt(S) + R2)*S
        Dens0 = SIG + DR350 + 1000.
        return Dens0


    def skillscore_trajectory(self, lon_obs, lat_obs, time_obs, duration=None,
                              max_time_offset=timedelta(seconds=60), method='liu-weissberg',
                              **kwargs):
        '''Calculate skill-score comparing simulated and observed trajectories

        A skill score is calculated for each of the modelled trajectories,
        starting from the closest time/obs of the given modelled trajectory

        Parameters
        ----------
        lon_obs : array_like
        lat_obs : array_like
            The longitude and latitudes of the observed trajectory
        time_obs : array of datetime
            The time corresponding to the observed positions
        duration : timedelta
            If None, skill score is calculated for the full trajectory, from the start
            Otherwise, only this sub-part of the trajectory is used.
        max_time_offset : timedelta
            The maximum allowed time difference between observed and modelled time
            Default: 1 minute
        method : string 
            Currently available: 'liu-wessberg'
            To be implemented: 'darpa'
        **kwargs : further arguments for skillscore method (e.g. tolerance_threshold)

        Returns
        -------
        skillscore : array_like
            One value for each trajectory
        '''

        time_obs = np.array(time_obs)
        timesteps_obs = time_obs[1:] - time_obs[0:-1]
        if timesteps_obs.min() != timesteps_obs.max():
            raise ValueError('Time step of trajectory is variable: %s to %s' %
                             (timesteps_obs.min(), timesteps_obs.max()))
        if timesteps_obs.min() != self.time_step_output:
            raise ValueError('Time step of trajectory is %s, but time step of simulation is %s.'
                    % (timesteps_obs.min(), self.time_step_output))

        from bisect import bisect_left
        times = np.array(self.get_time_array()[0])
        skillscore = np.zeros(self.num_elements_total())
        for i in range(self.num_elements_total()):
            lon = self.history['lon'][i]
            lat = self.history['lat'][i]
            status = self.history['status'][i]
            lon_model = lon[status==0]
            lat_model = lat[status==0]
            time_model = times[status==0]

            indo = bisect_left(time_obs, time_model[0], hi=len(time_obs)-1)
            if time_obs[indo+1] - time_model[0] < time_obs[indo] - time_model[0]:
                indo = indo + 1
            # Check that modeled and observed (sub)trajectories start at same position
            np.testing.assert_almost_equal(lon_model[0], lon_obs[indo], 3)
            np.testing.assert_almost_equal(lat_model[0], lat_obs[indo], 3)

            time_offset = np.abs((time_model[0] - time_obs[indo]).total_seconds())
            if time_offset > max_time_offset.total_seconds():
                raise ValueError('Too large offset: %s' % time_offset)
            length = len(time_obs) - indo
            indo = slice(indo, indo+length)
            skillscore[i] = skillscore_liu_weissberg(lon_obs[indo], lat_obs[indo],
                    lon_model[0:length], lat_model[0:length], **kwargs)

        return skillscore

        

    def advect_ocean_current(self, factor=1):

        cdf = self.elements.current_drift_factor
        cdfmin = cdf.min()
        cdfmax = cdf.max()
        if cdfmin != 1 or cdfmax != 1:
            if cdfmin == cdfmax:
                logger.debug('Using currrent drift factor of %s' % cdf)
            else:
                logger.debug('Using currrent drift factor between %s and %s'
                             % (cdfmin, cdfmax))
        factor = factor*cdf
        # Runge-Kutta scheme
        if self.get_config('drift:advection_scheme')[0:11] == 'runge-kutta':
            x_vel = self.environment.x_sea_water_velocity
            y_vel = self.environment.y_sea_water_velocity

            # Find midpoint
            az = np.degrees(np.arctan2(x_vel, y_vel))
            speed = np.sqrt(x_vel*x_vel + y_vel*y_vel)
            dist = speed*self.time_step.total_seconds()*.5
            geod = pyproj.Geod(ellps='WGS84')
            mid_lon, mid_lat, dummy = geod.fwd(self.elements.lon,
                                               self.elements.lat,
                                               az, dist, radians=False)
            # Find current at midpoint, a half timestep later
            logger.debug('Runge-kutta, fetching half time-step later...')
            mid_env, profiles, missing = self.get_environment(
                ['x_sea_water_velocity', 'y_sea_water_velocity'],
                self.time + self.time_step/2,
                mid_lon, mid_lat, self.elements.z, profiles=None)
            if self.get_config('drift:advection_scheme') == 'runge-kutta4':
                logger.debug('Runge-kutta 4th order...')
                x_vel2 = mid_env['x_sea_water_velocity']
                y_vel2 = mid_env['y_sea_water_velocity']
                az2 = np.degrees(np.arctan2(x_vel2, y_vel2))
                speed2 = np.sqrt(x_vel2*x_vel2 + y_vel2*y_vel2)
                dist2 = speed2*self.time_step.total_seconds()*.5
                lon2, lat2, dummy = \
                    geod.fwd(self.elements.lon,
                             self.elements.lat,
                             az2, dist2, radians=False)
                env2, profiles, missing = self.get_environment(
                    ['x_sea_water_velocity', 'y_sea_water_velocity'],
                    self.time + self.time_step/2,
                    lon2, lat2, self.elements.z, profiles=None)
                # Third step
                x_vel3 = env2['x_sea_water_velocity']
                y_vel3 = env2['y_sea_water_velocity']
                az3 = np.degrees(np.arctan2(x_vel3, y_vel3))
                speed3 = np.sqrt(x_vel3*x_vel3 + y_vel3*y_vel3)
                dist3 = speed3*self.time_step.total_seconds()*.5
                lon3, lat3, dummy = \
                    geod.fwd(self.elements.lon,
                             self.elements.lat,
                             az3, dist3, radians=False)
                env3, profiles, missing = self.get_environment(
                    ['x_sea_water_velocity', 'y_sea_water_velocity'],
                    self.time + self.time_step,
                    lon3, lat3, self.elements.z, profiles=None)
                # Fourth step
                x_vel4 = env3['x_sea_water_velocity']
                y_vel4 = env3['y_sea_water_velocity']
                u4 = (x_vel + 2*x_vel2 + 2* x_vel3 + x_vel4)/6.0
                v4 = (y_vel + 2*y_vel2 + 2* y_vel3 + y_vel4)/6.0
                # Move particles using runge-kutta4 velocity
                self.update_positions(u4*factor, v4*factor)

            else:
                # Move particles using runge-kutta velocity
                self.update_positions(
                        factor*mid_env['x_sea_water_velocity'],
                        factor*mid_env['y_sea_water_velocity'])
        elif self.get_config('drift:advection_scheme') == 'euler':
            # Euler scheme
            self.update_positions(
                    factor*self.environment.x_sea_water_velocity,
                    factor*self.environment.y_sea_water_velocity)
        else:
            raise ValueError('Drift scheme not recognised: ' +
                             self.get_config('drift:advection_scheme'))

    def advect_with_sea_ice(self, factor=1):
        if hasattr(self.environment, 'sea_ice_x_velocity'):
            self.update_positions(
                    factor*self.environment.sea_ice_x_velocity,
                    factor*self.environment.sea_ice_y_velocity)
        else:
            if not hasattr(self.environment, 'x_sea_water_velocity'):
                logger.info('No sea ice velocity available')
                return
            # Sea ice velocity, rule-of-thumb from Nordam,
            # doi:10.1016/j.marpolbul.2019.01.019
            ice_velocity_x = self.environment.x_sea_water_velocity + \
                    0.015*self.environment.x_wind
            ice_velocity_y = self.environment.y_sea_water_velocity + \
                    0.015*self.environment.y_wind
            self.update_positions(
                    factor*ice_velocity_x,
                    factor*ice_velocity_y)

    def advect_wind(self, factor=1):
        # Elements at/near ocean surface (z>wind_drift_depth) are advected with given percentage
        # of wind speed. NB: Only basic Euler scheme is implemented

        # Make copy to prevent outside value to be modified
        wind_drift_factor = self.elements.wind_drift_factor.copy()

        # Convert wind_drift_factor to array
        if len(np.atleast_1d(wind_drift_factor)) == 1:
            wind_drift_factor = wind_drift_factor*np.ones(len(self.elements))
        # Convert z to array
        if not hasattr(self.elements, 'z'):
            z = np.zeros(len(self.elements))  # Assumed on surface, if no z
        else:
            z = self.elements.z
        if len(np.atleast_1d(z)) == 1:
            z = z*np.ones(len(self.elements))

        try:
            wind_drift_depth = self.get_config('drift:wind_drift_depth')
        except:
            wind_drift_depth = 0
        if wind_drift_depth == 0:
            surface_only = True
        else:
            wind_drift_depth = np.abs(wind_drift_depth)*np.ones(len(self.elements))
            surface_only = False

        surface = self.elements.z >= -wind_drift_depth
        if surface.sum() == 0:
            if surface_only is True:
                logger.debug('All elements are below surface, no wind-induced shear drift')
            else:
                logger.debug('All elements are below %fm, no wind-induced shear drift' % wind_drift_depth[0])
            return

        wdf = wind_drift_factor.copy()
        if surface_only is False:
            # linear decrease from surface down to wind_drift_depth
            wdf = wdf*(wind_drift_depth+self.elements.z)/wind_drift_depth
        wdf[~surface] = 0.0
        wdfmin = wdf[surface].min()
        wdfmax = wdf[surface].max()

        x_wind = self.environment.x_wind.copy()
        y_wind = self.environment.y_wind.copy()

        try:
            if self.get_config('drift:relative_wind') is True:
                # Use wind relative to (subtracted) ocean current
                x_wind = x_wind - self.environment.x_sea_water_velocity
                y_wind = y_wind - self.environment.y_sea_water_velocity
        except:
            pass

        speed = np.sqrt(x_wind[surface]*x_wind[surface] + y_wind[surface]*y_wind[surface])
        if wdf[surface].max() == 0:
            logger.debug('Wind drift factor is 0')
            return
        if speed.max() == 0:
            logger.debug('No wind for wind-sheared ocean drift')
            return

        speed = speed*wdf[surface]
        if surface_only is True:
            logger.debug('Advecting %s of %i elements at surface with '
                          'wind-sheared ocean current (%f m/s - %f m/s)'
                          % (np.sum(surface), self.num_elements_active(),
                             speed.min(), speed.max()))
        else:
            logger.debug('Advecting %s of %i elements above %.3fm with '
                          'wind-sheared ocean current (%f m/s - %f m/s)'
                          % (np.sum(surface), self.num_elements_active(),
                             wind_drift_depth[0],
                             speed.min(), speed.max()))

        self.update_positions(x_wind*wdf*factor, y_wind*wdf*factor)

    def stokes_drift(self, factor=1):

        if self.get_config('drift:stokes_drift') is False:
            logger.debug('Stokes drift not activated')
            return

        if np.max(np.array(
            self.environment.sea_surface_wave_stokes_drift_x_velocity+
            self.environment.sea_surface_wave_stokes_drift_y_velocity)) \
                == 0:
            logger.debug('No Stokes drift velocity available')
            return

        wave_height = self.significant_wave_height()
        wave_period = self.wave_period()
        if np.max(np.array(wave_height)) == 0:
            logger.debug('Stokes drift is available, but not Hs: using Hs=1 for Stokes profile')
            wave_height = 1
        if np.max(np.array(wave_period)) == 0:
            logger.debug('Stokes drift is available, but not Tp: using Tp=8 for Stokes profile')
            wave_period = 8

        stokes_u, stokes_v, s = stokes_drift_profile_breivik(
            self.environment.sea_surface_wave_stokes_drift_x_velocity,
            self.environment.sea_surface_wave_stokes_drift_y_velocity,
            wave_height, wave_period, self.elements.z)

        self.update_positions(stokes_u*factor, stokes_v*factor)
        if s.min() == s.max():
            logger.debug('Advecting with Stokes drift (%s m/s)' % s.min())
        else:
            logger.debug('Advecting with Stokes drift (%s to %s m/s)' %
                      (s.min(), s.max()))

    def wave_stokes_drift_parameterised(self, wind, fetch):
        """
        Parameterise stokes drift based on pre calculated tables and fetch.
        """

        if not hasattr(self, 'stokes_coefficients'):
            Wf       = {}
            Wf['5000'] = (0.0173,0.0160,0.0152,0.0145,0.0139,0.0135,
                        0.0132,0.0129,0.0126,0.0124,0.0122,0.0121,
                        0.0119,0.0118,0.0117,0.0116,0.0114,0.0113,
                        0.0112,0.0112,0.0111,0.0110,0.0109,0.0109,
                        0.0108,0.0107,0.0106,0.0106,0.0106,0.0105)

            Wf['25000'] = (0.0173,0.0197,0.0201,0.0185,0.0181,0.0176,
                         0.0171,0.0167,0.0164,0.0160,0.0158,0.0155,
                         0.0153,0.0151,0.0149,0.0147,0.0146,0.0144,
                         0.0143,0.0142,0.0140,0.0139,0.0138,0.0137,
                         0.0136,0.0135,0.0135,0.0134,0.0133,0.0132)

            Wf['50000'] = (0.0173,0.0197,0.0210,0.0216,0.0201,0.0194,
                         0.0190,0.0186,0.0183,0.0179,0.0176,0.0173,
                         0.0171,0.0168,0.0166,0.0164,0.0162,0.0160,
                         0.0159,0.0157,0.0156,0.0155,0.0153,0.0152,
                         0.0151,0.0150,0.0149,0.0148,0.0147,0.0146)
            n = {'5000': 3, '25000':6, '50000': 6}  # Polynom order

            self.stokes_coefficients = {}
            for f in ['5000', '25000', '50000']:
                self.stokes_coefficients[f] = np.polyfit(
                    range(len(Wf[f])), Wf[f], n[f])

        windspeed = np.sqrt(wind[0]**2 + wind[1]**2)
        windspeed[windspeed>30] = 30
        wf = np.polyval(self.stokes_coefficients[fetch], windspeed)
        stokes_drift_x_velocity = wind[0]*wf
        stokes_drift_y_velocity = wind[1]*wf

        return stokes_drift_x_velocity, stokes_drift_y_velocity


    def wave_significant_height_parameterised(self, wind, fetch):
        """
        Parameterise significant wave height based on pre calculated tables and fetch.
        """
        if not hasattr(self, 'hs_coefficients'):
            Sw       = {}
            Sw['5000'] = (0.030,0.077,0.124,0.170,0.216,0.263,
                        0.311,0.360,0.409,0.459,0.509,0.560,
                        0.612,0.664,0.716,0.771,0.823,0.876,
                        0.932,0.987,1.041,1.095,1.152,1.210,
                        1.265,1.319,1.375,1.434,1.494,1.552)

            Sw['25000'] = (0.030,0.122,0.251,0.336,0.442,0.546,
                         0.650,0.753,0.856,0.959,1.063,1.168,
                         1.273,1.379,1.486,1.593,1.702,1.811,
                         1.920,2.030,2.142,2.254,2.366,2.478,
                         2.592,2.707,2.822,2.936,3.051,3.166)

            Sw['50000'] = (0.030,0.122,0.274,0.474,0.591,0.724,
                         0.873,1.021,1.168,1.314,1.460,1.606,
                         1.752,1.898,2.045,2.192,2.340,2.489,
                         2.639,2.789,2.940,3.092,3.244,3.397,
                         3.551,3.706,3.862,4.017,4.173,4.330)

            self.hs_coefficients = {}
            for f in ['5000', '25000', '50000']:
                self.hs_coefficients[f] = np.polyfit(
                    range(len(Sw[f])), Sw[f], 1)

        windspeed = np.sqrt(wind[0]**2 + wind[1]**2)
        windspeed[windspeed>30] = 30
        wave_significant_height = np.polyval(
            self.hs_coefficients[fetch], windspeed)

        return wave_significant_height

    def resurface_elements(self, minimum_depth):
        # Keep surfacing elements in water column as default,
        # i.e. no formation of surface slick
        surface = np.where(self.elements.z >= 0)[0]
        self.elements.z[surface] = minimum_depth

    def calculate_missing_environment_variables(self):

        # Missing significant wave height
        if hasattr(self.environment,
                   'sea_surface_wave_significant_height') and \
                self.environment.sea_surface_wave_significant_height.max() == 0:
            Hs = self.significant_wave_height()
            logger.debug('Calculating Hs from wind, min: %f, mean: %f, max: %f' %
                          (Hs.min(), Hs.mean(), Hs.max()))

        # Missing wave periode
        if hasattr(self.environment,
                   'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment') and \
                self.environment.sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment.max() == 0:
            wave_period = self.wave_period()
            logger.debug('Calculating wave period from wind, min: %f, mean: %f, max: %f' %
                          (wave_period.min(), wave_period.mean(), wave_period.max()))


    def wind_speed(self):
        return np.sqrt(self.environment.x_wind**2 +
                       self.environment.y_wind**2)

    def current_speed(self):
        return np.sqrt(self.environment.x_sea_water_velocity**2 +
                       self.environment.y_sea_water_velocity**2)

    def significant_wave_height(self):
        # Significant wave height, parameterise from wind if not available
        if hasattr(self.environment,
                   'sea_surface_wave_significant_height') and \
                self.environment.sea_surface_wave_significant_height.max() > 0:
            Hs = self.environment.sea_surface_wave_significant_height
        else:
            ## Neumann and Pierson, 1966
            #return 0.2*np.power(self.wind_speed(), 2)/9.81
            # WMO 1998
            Hs = 0.0246*np.power(self.wind_speed(), 2)
            self.environment.sea_surface_wave_significant_height = Hs

        return Hs

    def _wave_frequency(self):
        # Note: this is angular frequency, 2*pi*fp
        # Pierson-Moskowitz if period not available from readers
        # WMO guide to wave analysis and forecasting pp. 14, WMO (1998)
        windspeed = self.wind_speed()
        # fallback value if no wind speed or Hs to avoid division by zero
        omega = 5*np.ones(windspeed.shape)
        omega[windspeed>0] = 0.877*9.81/(1.17*windspeed[windspeed>0])
        return omega

    def wave_period(self):
        if hasattr(self.environment, 'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment'
                ) and self.environment.sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment.max() > 0:
            # prefer using Tm02:
            T = self.environment.sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment.copy()
            logger.debug('Using mean period Tm02 as wave period')
        elif hasattr(self.environment, 'sea_surface_wave_period_at_variance_spectral_density_maximum'
                ) and self.environment.sea_surface_wave_period_at_variance_spectral_density_maximum.max() > 0:
            # alternatively use Tp
            T = self.environment.sea_surface_wave_period_at_variance_spectral_density_maximum.copy()
            logger.debug('Using peak period Tp as wave period')
        else:
            # calculate Tp from wind speed:
            logger.debug('Calculating wave period Tm02 from wind')
            T = (2*np.pi)/self._wave_frequency()
            self.environment.sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment = T

        #print '\n T %s \n' % str(T.mean())
        if T.min() == 0:
            logger.warning('Zero wave period found - '
                            'replacing with mean')
            T[T==0] = np.mean(T[T>0])

        logger.debug('   min: %f, mean: %f, max: %f' % (T.min(), T.mean(), T.max()))

        return T

    def wave_energy(self):
        return 9.81*1028*np.power(self.significant_wave_height(), 2)/16

    def wave_energy_dissipation(self):
        # Delvigne and Sweeney
        return 0.0034*self.sea_water_density()*9.81 * \
            np.power(self.significant_wave_height(), 2)

    def wave_damping_coefficient(self):
        omega = 2*np.pi / self.wave_period()
        return (10E-5)*omega * \
            np.power(self.wave_energy(), 0.25)

    # def sea_water_density(self):
    #    return 1027  # kg/m3

    def sea_surface_wave_breaking_fraction(self):
        # TODO: We should also have an option here for
        # the case when wave height is given, but no wind
        f = 0.032*(self.wind_speed() - 5)/self.wave_period()
        f[f < 0] = 0
        return f

    def air_density(self):
        return 1.225  # Could calculate from temperature

    def windspeed_from_stress(self):
        wind_stress = np.sqrt(
            self.environment.surface_downward_x_stress**2 +
            self.environment.surface_downward_y_stress**2)
        return windspeed_from_stress_polyfit(wind_stress)

    def solar_elevation(self):
        '''Solar elevation at present time and position of active elements.'''
        return solar_elevation(self.time, self.elements.lon, self.elements.lat)

    def sea_floor_depth(self):
        '''Sea floor depth (positive) for presently active elements'''

        if hasattr(self, 'environment') and \
                hasattr(self.environment, 'sea_floor_depth_below_sea_level'):
            if len(self.environment.sea_floor_depth_below_sea_level) == \
                    self.num_elements_active():
                sea_floor_depth = \
                    self.environment.sea_floor_depth_below_sea_level
        if 'sea_floor_depth' not in locals():
            env, env_profiles, missing = \
                self.get_environment(['sea_floor_depth_below_sea_level'],
                                     time=self.time, lon=self.elements.lon,
                                     lat=self.elements.lat,
                                     z=0*self.elements.lon, profiles=None)
            sea_floor_depth = \
                env['sea_floor_depth_below_sea_level'].astype('float32')
        return sea_floor_depth


def wind_drag_coefficient(windspeed):
    '''Large and Pond (1981), J. Phys. Oceanog., 11, 324-336.'''
    Cd = 0.0012*np.ones(len(windspeed))
    Cd[windspeed > 11] = 0.001*(0.49 + 0.065*windspeed[windspeed > 11])
    return Cd

def windspeed_from_stress_polyfit(wind_stress):
    '''Inverting Large and Pond (1981) using polyfit'''
    windspeed = np.linspace(0, 30, 30)
    rho_air = 1.225
    stress = wind_drag_coefficient(windspeed)*rho_air*(windspeed**2)
    z = np.polyfit(stress, windspeed, 3)
    p = np.poly1d(z)
    return p(wind_stress)


def declination(time):
    '''Solar declination in degrees.'''
    try:
        day_of_year = time.timetuple().tm_yday
    except:
        day_of_year = np.asarray([t.timetuple().tm_yday for t in time])
    declination = \
        np.arcsin(np.deg2rad(-23.44)*
                  np.cos(np.radians((360.0/365.24)*(day_of_year + 10) +
                                    (360.0/np.pi)*0.0167*
                                    np.sin(np.radians((360.0/365.24)*
                                           (day_of_year - 2)))
                                    )))
    return np.rad2deg(declination)


def equation_of_time(time):
    '''Equation of time in minutes.'''
    time = np.atleast_1d(time)
    day_of_year = np.asarray([t.timetuple().tm_yday for t in time])
    hour = np.asarray([t.hour for t in time])
    gamma = 2*np.pi/365.0*(day_of_year - 1. +
                           (hour - 12.) / 24.)
    eqtime = 229.18 * (0.000075 + 0.001868*np.cos(gamma) -
                       0.032077*np.sin(gamma) -
                       0.014615*np.cos(2*gamma) - 0.040849*np.sin(2*gamma))
    return eqtime


def hour_angle(time, longitude):
    '''Solar hour angle in degrees.'''
    time = np.atleast_1d(time)
    time_offset = equation_of_time(time) + 4*longitude
    day_minutes = [t.hour*60.0 + t.minute + t.second/60.0 for t in time]
    true_solar_time = day_minutes + time_offset
    hour_angle = (true_solar_time/4.0) - 180.0  # degrees
    return hour_angle


def solar_elevation(time, longitude, latitude):
    '''Solar elevation in degrees.'''
    d_rad = np.deg2rad(declination(time))
    h = hour_angle(time, longitude)
    solar_elevation = np.rad2deg(np.arcsin(
        np.sin(np.deg2rad(latitude))*np.sin(d_rad) +
        np.cos(np.deg2rad(latitude))*np.cos(d_rad)*np.cos(np.deg2rad(h))))
    return solar_elevation



#
# Model class definitions for embedded content in our oil records
#
# Note: These are deprecated a bit.  An updated SARAFraction is sitting in
#       models.oil
from dataclasses import dataclass

from ..common.utilities import dataclass_to_json

from ..common.measurement import MassFraction


@dataclass_to_json
@dataclass
class Sara:
    method: str = None

    saturates: MassFraction = None
    aromatics: MassFraction = None
    resins: MassFraction = None
    asphaltenes: MassFraction = None



"""
Tools for helping make our data models.

So far: making dataclasses read/writable as JSON
"""


def something(val):
    '''
        much like python's "Truthy" and Falsey", but we want some values
        to not be false like zero, for instance
    '''
    return ((val == 0) or (val is not None) and val)


def dataclass_to_json(cls):
    """
    class decorator that adds the ability to save a dataclass as JSON

    All fields must be either JSON-able Python types or
    have be a type with a _to_json method
    """

    @classmethod
    def from_py_json(cls, py_json, allow_none=False):
        """
        classmethod to create a dataclass from json compatible python data
        structure.
        """
        # if there is a pre-processor - run it
        if hasattr(cls, "_pre_from_py_json"):
            py_json = cls._pre_from_py_json(py_json)

        arg_dict = {}

        if py_json is None and allow_none is True:
            # the parent object defined an attribute with a default of None
            # We could actually allow other default types, but this one is
            # common
            return py_json

        for fieldname, fieldobj in cls.__dataclass_fields__.items():
            if fieldname in py_json:
                allow_none = True if fieldobj.default is None else False

                try:  # see if it's "one of ours"
                    arg_dict[fieldname] = (fieldobj.type
                                           .from_py_json(py_json[fieldname],
                                                         allow_none=allow_none)
                                           )
                except AttributeError:
                    # it's not, so we just use the value
                    arg_dict[fieldname] = py_json[fieldname]
                except TypeError as err:
                    raise TypeError(f'TypeError in '
                                    f'{cls.__name__}._from_py_json(): '
                                    f'field: {fieldname}\n'
                                    f'{err.args}')

        obj = cls(**arg_dict)
        return obj

    def py_json(self, sparse=True):
        """
        function to convert a dataclass to json compatible python

        :param sparse=True: If sparse is True, only non-empty fields will be
                            written. If False, then all fields will be
                            included.
        """
        json_obj = {}
        for fieldname in self.__dataclass_fields__.keys():
            val = getattr(self, fieldname)
            try:  # convert to json
                val = val.py_json(sparse=sparse)
            except AttributeError:
                pass

            if not sparse:
                json_obj[fieldname] = val
            elif something(val):
                json_obj[fieldname] = val

        return json_obj

    def validate(self):
        """
        Function to validate a dataclass with fields that have validate
        methods.  The validate methods are expected to return a list of
        validation messages.

        The top-level validator extends the existing list
        """
        # I have NO idea how this happens, but it does ?!?!?
        if self is None:
            return []

        # first see if there is a "private" one:
        if hasattr(self, '_validate'):
            messages = self._validate()
        else:
            messages = []

        for fieldname, fieldobj in self.__dataclass_fields__.items():
            value = getattr(self, fieldname)
            if hasattr(fieldobj, 'type') and hasattr(fieldobj.type, 'validate'):
                messages.extend(fieldobj.type.validate(value))
            # try:
            #     # validate with the type's validate method
            #     messages.extend(fieldobj.type.validate(value))
            # except AttributeError as err:  # This one doesn't have a validate method.
            #     print("\nAttributeError:", err)
            #     pass

        return messages

    def __setattr__(self, name, val):
        try:
            _fieldobj = self.__dataclass_fields__[name]
        except KeyError:
            raise AttributeError(f"You can only set existing attributes: "
                                 f"{self.__class__.__name__}.{name} "
                                 "does not exist")
        self.__dict__[name] = val

    def __repr__(self):
        atts = ((att, getattr(self, att)) for att in self.__dataclass_fields__.keys())
        atts = (f'{att}={val!r}' for att, val in atts if val is not None)
        return f'{self.__class__.__name__}({", ".join(atts)})'

    cls.py_json = py_json

    cls.from_py_json = from_py_json

    if hasattr(cls, "validate"):
        cls._validate = cls.validate
    cls.validate = validate

    cls.__setattr__ = __setattr__
    cls.__repr__ = __repr__

    return cls


class JSON_List(list):
    """
    just like a list, but with the ability to turn it into JSON

    A regular list can only be converted to JSON if it has
    JSON-able objects in it.

    Note: must be subclassed, and the item_type attribute set
    """
    item_type = None

    def py_json(self, sparse=True):
        json_obj = []
        for item in self:
            try:
                json_obj.append(item.py_json(sparse))
            except AttributeError:
                json_obj.append(item)

        return json_obj

    @classmethod
    def from_py_json(cls, py_json, allow_none=False):
        """
        create a JSON_List from json array of objects
        that may be json-able.
        """
        if cls.item_type is None:
            raise TypeError("You can not reconstruct a list of unknown type")

        jl = cls()  # an empty JSON_List

        # loop through contents
        for item in py_json:
            jl.append(cls.item_type.from_py_json(item))

        return jl

    def __repr__(self):
        return f"{self.__class__.__name__}({list.__repr__(self)})"



import logging
logger = logging.getLogger(__name__)
from .grid import RegularGrid
from . import srs
from .interp import vec_nearest
from abc import abstractmethod
import numpy as np
from datetime import timedelta, datetime


class Simulation:
    r"""
    The simulation.

    It contains the problem to be simulated, means to read necessary input
    variables, and the physics for modeling the convection of the initial
    conditions.

    Convection:

      Convection consists of advection and diffusion.

      Diffusion is given by:

      .. math::

        \frac{\partial U}{\partial t} = D \left( \frac{\partial^2 U}{\partial x^2} + \frac{\partial^2 U}{\partial y^2} \right)


      The convection equation is (`wiki
      <https://en.wikipedia.org/wiki/Convection%E2%80%93diffusion_equation>`_):

      .. math::

        \frac{\partial c}{\partial t} = ...

      with the assumptions that:

        * the diffusion constant `D` is constant for the field,
        * and that the flow `u` is incompressible (i.e. has *no divergence*).

      the equation simplifies to:

      .. math::

        \frac{\partial c}{\partial t} = D \nabla^2 c - \mathbf{v} \cdot \nabla T

      where :math:`\nabla^2 = \triangle` is the Laplacian.
    """

    """The model grid."""
    grid = None

    """Reference time `datetime.Datetime`"""
    t0 = None

    """Current time offset after `t0`"""
    t = 0.

    """
    Diffusivity (:math:`m^2/s`). E.g. between 0.01 and 0.1 for oil on the
    surface of the ocean (`Matsuzakia et. al., 2017
    <https://www.sciencedirect.com/science/article/pii/S0025326X16308426>`_).

    Decreasing diffusivity places stricter stability criteria on time step.
    """
    D = 0.1

    """Porosity, rate of liquid volume to total volume (fraction of flux)"""
    rho = 1.

    """List of readers"""
    readers = None

    def __init__(self, grid):
        self.grid = grid
        self.readers = []
        self.t0 = datetime.utcnow()

    @classmethod
    def new(cls, lon0 = 10., lat0 = 65., res = 10., shape=(100, 100)):
        """
        New simulation on a :class:`.grid.RegularGrid`.

        Args:

            lon0: lower-left corner longitude

            lat0: lower-left corner latitude

            res: resolution (dx and dy)

            shape: shape (size) of grid
        """
        s = cls(grid=RegularGrid.new(lon0, lat0, res, shape))
        logger.info("New %s with Regular Grid" % (cls.__name__))

        return s

    def source(self, lon, lat, X):
        """
        Source `X` onto grid with lower-left corner `lon`, `lat`.
        """
        x0, y0 = self.grid.srs(lon, lat)
        assert self.grid.contains(x0, y0)

        ix0 = vec_nearest(self.grid.x, x0)
        iy0 = vec_nearest(self.grid.y, y0)
        print(ix0)
        ix1 = ix0 + X.shape[0]
        iy1 = iy0 + X.shape[1]

        assert ix1 < self.grid.grid.shape[0]
        assert iy1 < self.grid.grid.shape[1]

        self.grid.grid[ix0:ix1, iy0:iy1] = X

    def source_gaussian_blob(self, lon, lat, A=1., N=10, sigma=10.):
        r"""
        Source a Gaussian blob (2D normal distribution) at `lon` and `lat` with
        `sigma` (standard deviation, meters) radius.

        Args:

          lon, lat: Center coordinates, or :math:`\bar \mu`.

          A: Amplitude.

          N: Kernel size.

          sigma: standard deviation (:math:`\sigma`) in meters.
        """
        def gaussian2d(n, std):
            """
            Gaussian 2D window with size `n` and standard deviation `std`.
            """

            from scipy import signal
            gk = signal.windows.gaussian(n, std=std)
            return np.outer(gk, gk)

        # center of kernel
        x, y = self.grid.srs(lon, lat)
        assert self.grid.contains(x, y)

        ix0 = vec_nearest(self.grid.x, x) - N // 2
        iy0 = vec_nearest(self.grid.y, y) - N // 2
        ix1 = ix0 + N
        iy1 = iy0 + N

        assert ix0 > 0 and iy0 > 0

        S = gaussian2d(N, sigma / self.grid.res)

        self.grid.grid[ix0:ix1, iy0:iy1] = S

    def U(self, t):
        """
        Get U (ocean current) for t0 + t
        """
        Ux, Uy = None, None
        t = self.t0 + timedelta(seconds = t)

        for r in self.readers:
            vv = r.variables()
            if 'x_sea_water_velocity' in vv and 'y_sea_water_velocity' in vv:
                Ux, Uy = r.read_grid(self.grid, ['x_sea_water_velocity', 'y_sea_water_velocity'], t)

            if Ux is not None and Uy is not None:
                break

        return Ux, Uy

    @abstractmethod
    def step(self, dt=None):
        """
        Step the simulation.

        Stepping the simulation involves applying diffusion and advection to the field.

        Args:

          dt: time delta (or use automatic).
        """
        pass

    @abstractmethod
    def integrate(self, dt=None, max_t=None, max_steps=None, observer = None):
        """
        Run simulation until termination condition is met.

        Args:

          dt: override time step

          observer: function to call after each step taking simulation object as first argument. the function may return False to stop the integration.

          Termination conditions:

          max_t: max time
          max_steps: max iterations
        """

        if max_t is not None and max_steps is not None:
            logger.warning(
                "no termination condition supplied, using max_steps = 1000")
            max_steps = 1000

        logger.info("integrating.. (dt = %s)", dt)

        n = 0

        while True:
            logger.debug("step: n = %d, t = %s (dt = %s)" % (n, self.t, dt))

            if max_t is not None and self.t >= max_t:
                logger.info("max_t reached.")
                break

            if max_steps is not None and n >= max_steps:
                logger.info("max_steps reached.")
                break

            self.step(dt=dt)

            n += 1


class ExplSimulation(Simulation):
    r"""
    A simple explicit scheme for integrating the convection-equation.

    * Forward difference in time
    * `ndimage.laplace` and `np.gradient` for spatial differences.

    https://en.wikipedia.org/wiki/Numerical_solution_of_the_convection%E2%80%93diffusion_equation#Solving_the_convection%E2%80%93diffusion_equation_using_the_finite_difference_method

    .. seealso::

      :class:`Simulation`.

    """
    def stability(self, dx, D, umax):
        """
        https://en.wikipedia.org/wiki/Numerical_solution_of_the_convection%E2%80%93diffusion_equation#Solving_the_convection%E2%80%93diffusion_equation_using_the_finite_difference_method
        """
        h = 2 * D / (self.rho * umax)
        dt = dx**2 / (2 * D)

        return h, dt

    def step(self, dt=None):
        from scipy import ndimage

        dx, dy = self.grid.res, self.grid.res
        D = self.D
        # Ux = np.sqrt(50 * 1.) * np.ones(self.grid.grid.shape)
        # Uy = Ux
        Ux, Uy = self.U(self.t)
        maxu = np.max(np.sqrt(Ux**2 + Uy**2).ravel())
        logger.debug("maxu = %s" % maxu)

        h, ddt = self.stability(dx, D, maxu)

        if h < dx:
            logger.warning("dx too big, dx = %s > h = %s" % (dx, h))

        if dt is None:
            dt = ddt
        elif ddt < dt:
            logger.warning("dt too big, dt = %s > ddt = %s" % (dt, ddt))

        # diffusion
        diff = D * 1. / dx**2 * ndimage.laplace(self.grid.grid)

        # advection
        gfx, gfy = np.gradient(self.grid.grid, dx, dy)
        adv = -(gfx * Ux + gfy * Uy)

        self.grid.grid = self.grid.grid + dt * (diff + adv)

        self.t += dt



"""
dataclass to hold the ESTS Hydrocarbon Fractions

This is kind of a special case, so this nails it down
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json

from .compound import CompoundList


@dataclass_to_json
@dataclass
class ESTSFractions:
    """
    hold the ESTS Hydrocarbon Fractions from env CA

    (https://www.ccme.ca/en/resources/canadian_environmental_quality_guidelines/index.html)
    """
    method: str = None

    saturates: CompoundList = field(default_factory=CompoundList)
    aromatics: CompoundList = field(default_factory=CompoundList)
    GC_TPH: CompoundList = field(default_factory=CompoundList)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Gaute Hope, MET Norway
"""
Interface to the ADIOS oil database.
"""

import logging

logger = logging.getLogger(__name__)

from . import dirjs
from . import oil
from .dirjs import get_oil_names, oils, find_full_oil_from_name, get_full_oil_from_id

oil_name_alias = {'EKOFISK BLEND 2002': 'EKOFISK BLEND 2000'}



#!/usr/bin/env python

import matplotlib
from matplotlib import pyplot as plt
if __name__ == '__main__':
    matplotlib.use('TKAgg')

import sys
import os
from datetime import datetime, timedelta
import numpy as np

from PIL import ImageTk, Image
import tkinter as tk
from tkinter import ttk
import opendrift
from opendrift.models.oceandrift import OceanDrift
from opendrift.models.openoil import OpenOil
from opendrift.models.leeway import Leeway
from opendrift.models.shipdrift import ShipDrift
from opendrift.models.openberg import OpenBerg
from opendrift.models.plastdrift import PlastDrift
from opendrift.models.radionuclides import RadionuclideDrift

# Class to redirect output to text box
class TextRedirector:

    def __init__(self, widget, tag='stdout'):
        self.defstdout = sys.stdout
        self.widget = widget
        self.tag = tag

    def write(self, str):
        self.widget.configure(state='normal')
        self.widget.insert('end', str, (self.tag,))
        self.widget.update_idletasks()
        self.widget.see(tk.END)

    def flush(self):
        self.defstdout.flush()

# Class for help-text
# https://stackoverflow.com/questions/20399243/display-message-when-hovering-over-something-with-mouse-cursor-in-python
class ToolTip(object):

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 57
        y = y + cy + self.widget.winfo_rooty() +27
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(tw, text=self.text, justify=tk.LEFT,
                      background="#ffff00", relief=tk.SOLID, borderwidth=1,
                      font=("tahoma", "10", "normal"))
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()

def CreateToolTip(widget, text):
    toolTip = ToolTip(widget)
    def enter(event):
        toolTip.showtip(text)
    def leave(event):
        toolTip.hidetip()
    widget.bind('<Enter>', enter)
    widget.bind('<Leave>', leave)


class OpenDriftGUI(tk.Tk):

    # Supported models as dictionary {model_name:model_class}
    opendrift_models = {m.__name__:m for m in
        [Leeway, OpenOil, ShipDrift, OpenBerg, OceanDrift, PlastDrift, RadionuclideDrift]}

    extra_args = {'OpenOil': {'location': 'NORWAY'}}

    # Overriding some default config settings, suitable for GUI
    # TODO: should be set as default-default
    GUI_config = {
            'general:time_step_minutes': {'default': 15, 'min': 1},
            'general:time_step_output_minutes': {'default': 30, 'min': 5},
            'seed:number': {'default': 5000, 'max': 100000},
            'seed:m3_per_hour': {'default': 100}
            }

    def __init__(self):

        tk.Tk.__init__(self)

        self.title('OpenDrift ' + opendrift.__version__ + ' GTI Turbo Ultra')

        ##################
        # Layout frames
        ##################
        self.n = ttk.Notebook(self.master)
        self.n.grid()
        self.seed = ttk.Frame(self.n)
        self.confignotebook = ttk.Notebook(self.n)
        self.config = ttk.Frame(self.confignotebook)
        self.forcing = ttk.Frame(self.n)
        self.n.add(self.seed, text='Seeding')
        self.n.add(self.confignotebook, text='Config')
        self.n.add(self.forcing, text='Forcing')
        self.confignotebook.add(self.config, text='SubConfig')

        # Top
        self.top = tk.Frame(self.seed,
                            relief=tk.FLAT, pady=25, padx=25)
        self.top.grid(row=0, column=1, rowspan=1)
        # Time start and end
        self.start_t = tk.Frame(self.seed, relief=tk.FLAT)
        self.start_t.grid(row=20, column=0, rowspan=1)
        self.end_t = tk.Frame(self.seed, relief=tk.FLAT)
        self.end_t.grid(row=30, column=0, rowspan=1)
        self.start = tk.Frame(self.seed, bg='lightgray', bd=2,
                              relief=tk.SUNKEN, pady=5, padx=5)
        self.start.grid(row=20, column=1, rowspan=1)
        self.end = tk.Frame(self.seed, bg='gray', bd=2,
                            relief=tk.SUNKEN, padx=5, pady=5)
        self.end.grid(row=30, column=1)
        self.coastline = tk.Frame(self.seed, bd=2,
                                 relief=tk.FLAT, padx=5, pady=0)
        self.coastline.grid(row=40, column=1)
        self.duration = tk.Frame(self.seed, bd=2,
                                 relief=tk.FLAT, padx=5, pady=5)
        self.duration.grid(row=50, column=1)
        self.output = tk.Frame(self.seed, bd=2,
                               relief=tk.FLAT, padx=5, pady=0)
        self.output.grid(row=70, column=0, columnspan=7, sticky='nsew')

        self.results = tk.Frame(self.seed, bd=2,
                               relief=tk.FLAT, padx=5, pady=0)
        self.results.grid(row=60, column=7, columnspan=1, sticky='ew')

        #######################################################
        tk.Label(self.top, text='Simulation type').grid(row=0, column=0)
        self.model = tk.StringVar()
        self.model.set(list(self.opendrift_models)[0])
        self.modeldrop = tk.OptionMenu(self.top, self.model,
            *(list(self.opendrift_models)), command=self.set_model)
        self.modeldrop.grid(row=0, column=1)

        help_button = tk.Button(self.top, text='Help',
                                command=self.show_help)
        help_button.grid(row=0, column=2, padx=50)


        ##########
        # Release
        ##########
        startlabel = tk.Label(self.start_t, text="\n\nStart release\n\n")
        startlabel.grid(row=0, column=0)

        tk.Label(self.start, text='Longitude').grid(row=0, column=1)
        tk.Label(self.start, text='Latitude').grid(row=0, column=0)
        tk.Label(self.start, text='Radius [m]').grid(row=0, column=2)
        self.latvar = tk.StringVar()
        self.lonvar = tk.StringVar()
        self.radiusvar = tk.StringVar()
        self.lat = tk.Entry(self.start, textvariable=self.latvar,
                            width=10, justify=tk.RIGHT)
        self.lon = tk.Entry(self.start, textvariable=self.lonvar,
                            width=10, justify=tk.RIGHT)
        self.radius = tk.Entry(self.start, width=6,
                               textvariable=self.radiusvar,
                               justify=tk.RIGHT)
        self.lon.grid(row=10, column=1)
        self.lon.insert(0, '4.5')
        self.lat.grid(row=10, column=0)
        self.lat.insert(0, '60.0')
        self.radius.grid(row=10, column=2)
        self.radius.insert(0, '1000')
        self.lonvar.trace('w', self.copy_position)
        self.latvar.trace('w', self.copy_position)
        self.radiusvar.trace('w', self.copy_position)
        conv=tk.Label(self.start, text='Convert from deg/min/sec', fg='blue')
        conv.grid(row=11, column=0, columnspan=2)
        conv.bind("<Button-1>", lambda e: self.convert_lonlat())

        ##########
        # Time
        ##########
        now = datetime.utcnow()
        tk.Label(self.start, text='Day').grid(row=20, column=0)
        tk.Label(self.start, text='Month').grid(row=20, column=1)
        tk.Label(self.start, text='Year').grid(row=20, column=2)
        tk.Label(self.start, text='Hour').grid(row=20, column=3)
        tk.Label(self.start, text='Minutes [UTC]').grid(row=20, column=4)
        self.datevar = tk.StringVar()
        self.dates = range(1, 32)
        self.datevar.set(now.day)
        self.date = tk.OptionMenu(self.start, self.datevar, *self.dates)
        self.date.grid(row=30, column=0)

        self.monthvar = tk.StringVar()
        self.months = ['January', 'February', 'March', 'April', 'May',
                       'June', 'July', 'August', 'September', 'October',
                       'November', 'December']
        self.monthvar.set(self.months[now.month-1])
        self.month = tk.OptionMenu(self.start, self.monthvar,
                                   *self.months)
        self.month.grid(row=30, column=1)

        self.yearvar = tk.StringVar()
        self.years = range(2015, now.year+2)
        self.yearvar.set(now.year)
        self.year = tk.OptionMenu(self.start, self.yearvar, *self.years)
        self.year.grid(row=30, column=2)

        self.hourvar = tk.StringVar()
        self.hours = range(0, 24)
        self.hourvar.set(now.hour)
        self.hour = tk.OptionMenu(self.start, self.hourvar, *self.hours)
        self.hour.grid(row=30, column=3)

        self.minutevar = tk.StringVar()
        self.minutes = range(0, 60, 5)
        self.minutevar.set(now.minute)
        self.minute = tk.OptionMenu(self.start, self.minutevar,
                                    *self.minutes)
        self.minute.grid(row=30, column=4)

        self.datevar.trace('w', self.copy_position)
        self.monthvar.trace('w', self.copy_position)
        self.yearvar.trace('w', self.copy_position)
        self.hourvar.trace('w', self.copy_position)
        self.minutevar.trace('w', self.copy_position)

        ###############
        # Release End
        ###############
        endlabel = tk.Label(self.end_t, text="\n\nEnd release\n\n")
        endlabel.grid(row=0, column=0)
        tk.Label(self.end, text='Longitude', bg='gray').grid(row=0, column=1)
        tk.Label(self.end, text='Latitude', bg='gray').grid(row=0, column=0)
        tk.Label(self.end, text='Radius [m]', bg='gray').grid(row=0, column=2)
        self.elat = tk.Entry(self.end, width=10, justify=tk.RIGHT)
        self.elon = tk.Entry(self.end, width=10, justify=tk.RIGHT)
        self.eradius = tk.Entry(self.end, width=6, justify=tk.RIGHT)
        self.elon.grid(row=10, column=1)
        self.elon.insert(0, '4.5')
        self.elat.grid(row=10, column=0)
        self.elat.insert(0, '60.0')
        self.eradius.grid(row=10, column=2)
        self.eradius.insert(0, '1000')
        ##########
        # Time
        ##########
        now = datetime.utcnow()
        tk.Label(self.end, text='Day', bg='gray').grid(row=20, column=0)
        tk.Label(self.end, text='Month', bg='gray').grid(row=20, column=1)
        tk.Label(self.end, text='Year', bg='gray').grid(row=20, column=2)
        tk.Label(self.end, text='Hour', bg='gray').grid(row=20, column=3)
        tk.Label(self.end, text='Minutes [UTC]', bg='gray').grid(row=20, column=4)
        self.edatevar = tk.StringVar()
        self.edates = range(1, 32)
        self.edatevar.set(now.day)
        self.edate = tk.OptionMenu(self.end, self.edatevar, *self.edates)
        self.edate.grid(row=30, column=0)

        self.emonthvar = tk.StringVar()
        self.emonthvar.set(self.months[now.month-1])
        self.emonth = tk.OptionMenu(self.end, self.emonthvar,
                                    *self.months)
        self.emonth.grid(row=30, column=1)

        self.eyearvar = tk.StringVar()
        self.eyears = range(2015, now.year+2)
        self.eyearvar.set(now.year)
        self.eyear = tk.OptionMenu(self.end, self.eyearvar, *self.eyears)
        self.eyear.grid(row=30, column=2)

        self.ehourvar = tk.StringVar()
        self.ehours = range(0, 24)
        self.ehourvar.set(now.hour)
        self.ehour = tk.OptionMenu(self.end, self.ehourvar, *self.ehours)
        self.ehour.grid(row=30, column=3)

        self.eminutevar = tk.StringVar()
        self.eminutes = range(0, 60, 5)
        self.eminutevar.set(now.minute)
        self.eminute = tk.OptionMenu(self.end, self.eminutevar,
                                     *self.eminutes)
        self.eminute.grid(row=30, column=4)
        self.eyear.config(bg='gray')
        self.emonth.config(bg='gray')
        self.edate.config(bg='gray')
        self.ehour.config(bg='gray')
        self.eminute.config(bg='gray')

        # Check seeding
        check_seed = tk.Button(self.end_t, text='Check seeding',
                               command=self.check_seeding)
        check_seed.grid(row=10, column=0, padx=0)

        #######################
        # Simulation duration
        #######################
        tk.Label(self.duration, text='Run simulation ').grid(row=50, column=0)
        self.durationhours = tk.Entry(self.duration, width=3,
                                      justify=tk.RIGHT)
        self.durationhours.grid(row=50, column=1)
        self.durationhours.insert(0, 12)
        tk.Label(self.duration, text=' hours ').grid(row=50, column=2)

        self.directionvar = tk.StringVar()
        self.directionvar.set('forwards')
        self.direction = tk.OptionMenu(self.duration, self.directionvar,
                                       'forwards', 'backwards')
        self.direction.grid(row=50, column=3)
        tk.Label(self.duration, text=' in time ').grid(row=50, column=4)

        ##############
        # Output box
        ##############
        self.text = tk.Text(self.output, wrap="word", height=18)
        self.text.grid(row=60, columnspan=6, sticky='nsw')
        self.text.tag_configure("stderr", foreground="#b22222")
        sys.stdout = TextRedirector(self.text, "stdout")
        sys.stderr = TextRedirector(self.text, "stderr")
        s = tk.Scrollbar(self)
        s.grid(row=60, column=6, sticky='ns')
        s.config(command=self.text.yview)
        self.text.config(yscrollcommand=s.set)

        # Diana
        self.dianadir = '/vol/vvfelles/opendrift/output/'
        if os.path.exists(self.dianadir):
            self.has_diana = True
            print('Diana is available!')
            self.outputdir = '/vol/vvfelles/opendrift/output_native/'
            startbutton = 'PEIS PAO'
        else:
            self.has_diana = False
            startbutton = 'START'

        ##############
        # Initialise
        ##############
        self.set_model(list(self.opendrift_models)[0])

        forcingfiles = open(self.o.test_data_folder() + '../../opendrift/scripts/data_sources.txt').readlines()
        print(forcingfiles)
        for i, ff in enumerate(forcingfiles):
            tk.Label(self.forcing, text=ff.strip(), wraplength=650, font=('Courier', 8)).grid(
                     row=i, column=0, sticky=tk.W)

        ##########################
        try:
            img = ImageTk.PhotoImage(Image.open(
                self.o.test_data_folder() +
                                     '../../docs/opendrift_logo.png'))
            panel = tk.Label(self.seed, image=img)
            panel.image = img
            panel.grid(row=0, column=0)
        except Exception as e:
            print(e)
            pass # Could not display logo

        ##########
        # RUN
        ##########
        tk.Button(self.seed, text=startbutton, bg='green',
                  command=self.run_opendrift).grid(row=80, column=1,
                                              sticky=tk.W, pady=4)

    def copy_position(self, a, b, c):
        self.elat.delete(0, tk.END)
        self.elat.insert(0, self.lat.get())
        self.elon.delete(0, tk.END)
        self.elon.insert(0, self.lon.get())
        self.eradius.delete(0, tk.END)
        self.eradius.insert(0, self.radius.get())
        self.edatevar.set(self.datevar.get())
        self.emonthvar.set(self.monthvar.get())
        self.eyearvar.set(self.yearvar.get())
        self.ehourvar.set(self.hourvar.get())
        self.eminutevar.set(self.minutevar.get())

    def handle_result(self, command):

        from os.path import expanduser
        homefolder = expanduser("~")
        filename = homefolder + '/' + self.simulationname

        if command[0:4] == 'save':
            plt.switch_backend('agg')
        elif command[0:4] == 'show':
            plt.switch_backend('TkAgg')

        if command == 'saveanimation':
            filename = filename + '.mp4'
            self.o.animation(filename=filename)
            print('='*30 + '\nAnimation saved to file:\n'
                  + filename + '\n' + '='*30)
        elif command == 'showanimation':
            self.o.animation()
        elif command == 'saveplot':
            filename = filename + '.png'
            self.o.plot(filename=filename)
            print('='*30 + '\nPlot saved to file:\n'
                  + filename + '\n' + '='*30)
        elif command == 'showplot':
            self.o.plot()
        elif command == 'showoilbudget':
            self.o.plot_oil_budget()
        elif command == 'saveoilbudget':
            filename = filename + '_oilbudget.png'
            self.o.plot_oil_budget(filename=filename)
            print('='*30 + '\nPlot saved to file: '
                  + filename + '\n' + '='*30)

        elif command == 'showanimationspecie':
            self.o.animation(
                color='specie',vmin=0,vmax=self.o.nspecies-1,
                colorbar=False,
                legend=[self.o.specie_num2name(i) for i in range(self.o.nspecies)]
                )
        elif command == 'saveanimationspecie':
            filename = filename + '.mp4'
            self.o.animation(filename=filename,
                color='specie',vmin=0,vmax=self.o.nspecies-1,
                                colorbar=False,
                legend=[self.o.specie_num2name(i) for i in range(self.o.nspecies)]
                )
        elif command == 'saveconcfile':
            self.o.guipp_saveconcfile(filename=homefolder+'/conc_radio.nc')
        elif command == 'plotconc':
            zlayer = [-1]
            time   = None
            specie = ['LMM']
            self.o.guipp_plotandsaveconc(filename=homefolder+'/conc_radio.nc',
                                         outfilename=homefolder+'/radio_plots/RadioConc',
                                         zlayers=zlayer, time=time, specie=specie )
        elif command == 'showanimationprofile':
            self.o.guipp_showanimationprofile()



    def validate_config(self, value_if_allowed, prior_value, key):
        """From config menu selection."""
        print('Input: %s -> %s', (key, value_if_allowed))
        if value_if_allowed == 'None':
            print('Setting None value')
            return True
        if value_if_allowed in ' -':
            print('Allowing temporally empty or minus sign')
            return True
        sc = self.o._config[key]
        if sc['type'] in ['int', 'float']:
            try:
                value_if_allowed = float(value_if_allowed)
            except:
                print('Nonumber')
                return False
        try:
            print('Setting: %s -> %s', (key, value_if_allowed))
            self.o.set_config(key, value_if_allowed)
            return True
        except:
            return False

    def set_model(self, model, rebuild_gui=True):

        # Creating simulation object (self.o) of chosen model class
        print('Setting model: ' + model)
        if model in self.extra_args:
            extra_args = self.extra_args[model]
        else:
            extra_args = {}
        self.o = self.opendrift_models[model](**extra_args)
        self.modelname = model  # So that new instance may be initiated at repeated run

        # Setting GUI-specific default config values
        for k,v in self.GUI_config.items():
            try:
                if 'default' in v:
                    self.o._set_config_default(k, v['default'])
                if 'min' in v:
                    self.o._config[k]['min'] = v['min']
                if 'max' in v:
                    self.o._config[k]['max'] = v['max']
            except:
                pass

        if rebuild_gui is False:
            return

        # Remove current GUI components and rebuild with new
        for con in self.confignotebook.winfo_children():
            con.destroy()
        self.subconfig = {}
        confnames = list(set([cn.split(':')[0] for cn in self.o._config]))
        confnames.extend(['environment:constant', 'environment:fallback'])
        confnames.remove('environment')
        for sub in confnames:
            self.subconfig[sub] = tk.Frame(self.confignotebook, pady=25)
            self.confignotebook.add(self.subconfig[sub], text=sub)

        sc = self.o.get_configspec(level=[2, 3])
        self.config_input = {}
        self.config_input_var = {}
        for i, key in enumerate(list(sc)):
            if key.startswith('environment:constant'):
                tab = self.subconfig['environment:constant']
                keystr = key.split(':')[-1]
            elif key.startswith('environment:fallback'):
                tab = self.subconfig['environment:fallback']
                keystr = key.split(':')[-1]
            else:
                tab = self.subconfig[key.split(':')[0]]
                keystr = ''.join(key.split(':')[1:])

            lab = tk.Label(tab, text=keystr)
            lab.grid(row=i, column=1, rowspan=1)
            if sc[key]['type'] in ['float', 'int']:
                self.config_input_var[i] = tk.StringVar()
                vcmd = (tab.register(self.validate_config),
                    '%P', '%s', key)
                self.config_input[i] = tk.Entry(
                    tab, textvariable=self.config_input_var[i],
                    validate='key', validatecommand=vcmd,
                    width=12, justify=tk.RIGHT)
                self.config_input[i].insert(0, str(sc[key]['default']))
                self.config_input[i].grid(row=i, column=2, rowspan=1)
                tk.Label(tab, text='[%s]  min: %s, max: %s' % (
                    sc[key]['units'], sc[key]['min'], sc[key]['max'])
                        ).grid(row=i, column=3, rowspan=1)
            elif sc[key]['type'] == 'bool':
                if self.o.get_config(key) is True:
                    value = 1
                else:
                    value = 0
                self.config_input_var[i] = tk.IntVar(value=value)
                vcb = (tab.register(self.set_config_checkbox),
                       key, i)
                self.config_input[i] = tk.Checkbutton(
                    tab, variable=self.config_input_var[i],
                    command=vcb, text='')
                self.config_input[i].grid(row=i, column=2, rowspan=1)
            elif sc[key]['type'] == 'enum':
                self.config_input_var[i] = tk.StringVar(value=self.o.get_config(key))
                width = len(max(sc[key]['enum'], key=len))
                self.config_input[i] = ttk.Combobox(
                    tab, width=width,
                    textvariable=self.config_input_var[i],
                    values=sc[key]['enum'])
                self.config_input[i].bind("<<ComboboxSelected>>",
                        lambda event, keyx=key, ix=i:
                            self.set_config_enum(event, keyx, ix))
                self.config_input[i].grid(row=i, column=2, rowspan=1)

            CreateToolTip(lab, sc[key]['description'])

        try:
            self.results.destroy()
        except:
            pass

        # Only ESSENTIAL config items are shown on front page with seeding
        sc = self.o.get_configspec(level=self.o.CONFIG_LEVEL_ESSENTIAL)
        self.seed_input = {}
        self.seed_input_var = {}
        self.seed_input_label = {}
        self.seed_frame = tk.Frame(self.seed, bd=2,
                                   relief=tk.FLAT, padx=5, pady=0)
        self.seed_frame.grid(row=60, columnspan=8, sticky='nsew')
        # FIND
        for num, i in enumerate(sc):
            varlabel = i.split(':')[-1]
            if i in self.o.ElementType.variables.keys():
                if 'units' in self.o.ElementType.variables[i].keys():
                    units = self.o.ElementType.variables[i]['units']
                    if units == '1':
                        units = 'fraction'
                    varlabel = '%s [%s]' % (varlabel, units)

            self.seed_input_label[i] = tk.Label(self.seed_frame,
                                                text=varlabel + '\t')
            self.seed_input_label[i].grid(row=num, column=0)
            CreateToolTip(self.seed_input_label[i], text=sc[i]['description'])
            actual_val = self.o.get_config(i)
            if sc[i]['type'] == 'enum':
                self.seed_input_var[i] = tk.StringVar()
                self.seed_input[i] = ttk.Combobox(
                    self.seed_frame, width=50,
                    textvariable=self.seed_input_var[i],
                    values=sc[i]['enum'])
                self.seed_input_var[i].set(actual_val)
            elif sc[i]['type'] == 'bool':
                self.seed_input_var[i] = tk.IntVar(value=sc[i]['value'])
                self.seed_input[i] = tk.Checkbutton(
                    self.seed_frame, variable=self.seed_input_var[i],
                    text=sc[i]['description'])
            else:
                self.seed_input_var[i] = tk.StringVar()
                self.seed_input[i] = tk.Entry(
                    self.seed_frame, textvariable=self.seed_input_var[i],
                    width=12, justify=tk.RIGHT)
                self.seed_input[i].insert(0, actual_val)
            self.seed_input[i].grid(row=num, column=1)

    def set_config_checkbox(self, key, i):
        i = int(i)
        newval = self.config_input_var[i].get()
        if newval == 0:
            print('Setting %s to False' % key)
            self.o.set_config(key, False)
        elif newval == 1:
            print('Setting %s to True' % key)
            self.o.set_config(key, True)

    def set_config_enum(self, event, key, i):
        newval = self.config_input_var[i].get()
        print('Setting ' + key + newval)
        self.o.set_config(key, newval)

    def show_help(self):
        help_url = 'https://opendrift.github.io/gui.html'
        print('Opening help website:\n' + help_url)
        import webbrowser
        webbrowser.open(help_url)

    def convert_lonlat(self):
        convert_url = 'https://www.rapidtables.com/convert/number/degrees-minutes-seconds-to-degrees.html'
        print('Opening conversion website:\n' + convert_url)
        import webbrowser
        webbrowser.open(convert_url)

    def check_seeding(self):
        print('#'*50)
        print('Hang on, plot is comming in a few seconds...')
        print('#'*50)
        month = int(self.months.index(self.monthvar.get()) + 1)
        start_time = datetime(int(self.yearvar.get()), month,
                              int(self.datevar.get()),
                              int(self.hourvar.get()),
                              int(self.minutevar.get()))
        emonth = int(self.months.index(self.emonthvar.get()) + 1)
        end_time = datetime(int(self.eyearvar.get()), emonth,
                            int(self.edatevar.get()),
                            int(self.ehourvar.get()),
                            int(self.eminutevar.get()))
        sys.stdout.flush()
        lon = float(self.lon.get())
        lat = float(self.lat.get())
        radius = float(self.radius.get())
        elon = float(self.elon.get())
        elat = float(self.elat.get())
        eradius = float(self.eradius.get())
        if lon != elon or lat != elat or start_time != end_time:
            lon = [lon, elon]
            lat = [lat, elat]
            radius = [radius, eradius]
            start_time = [start_time, end_time]
            cone = True
        else:
            cone = False

        so = Leeway(loglevel=50)
        for k,v in self.GUI_config.items():
            try:
                so.set_config(k, v)
            except:
                pass
        so.seed_cone(lon=lon, lat=lat, radius=radius, time=start_time)
        so.plot(buffer=.5, fast=True)
        del so

    def run_opendrift(self):
        sys.stdout.write('running OpenDrift')

        # Creating fresh instance of the current model, but keeping config
        adjusted_config = self.o._config
        self.set_model(self.modelname, rebuild_gui=False)
        self.o._config = adjusted_config

        try:
            self.budgetbutton.destroy()
        except Exception as e:
            print(e)
            pass
        month = int(self.months.index(self.monthvar.get()) + 1)
        start_time = datetime(int(self.yearvar.get()), month,
                              int(self.datevar.get()),
                              int(self.hourvar.get()),
                              int(self.minutevar.get()))
        emonth = int(self.months.index(self.emonthvar.get()) + 1)
        end_time = datetime(int(self.eyearvar.get()), emonth,
                            int(self.edatevar.get()),
                            int(self.ehourvar.get()),
                            int(self.eminutevar.get()))
        sys.stdout.flush()
        lon = float(self.lon.get())
        lat = float(self.lat.get())
        radius = float(self.radius.get())
        elon = float(self.elon.get())
        elat = float(self.elat.get())
        eradius = float(self.eradius.get())
        if lon != elon or lat != elat or start_time != end_time:
            lon = [lon, elon]
            lat = [lat, elat]
            radius = [radius, eradius]
            start_time = [start_time, end_time]
            cone = True
        else:
            cone = False

        for se in self.seed_input:
            val = self.seed_input_var[se].get()
            if self.o._config[se]['type'] in ['float', 'int']:
                val = float(val)
            elif self.o._config[se]['type'] == 'bool':
                if val == 1:
                    val = True
                elif val == 0:
                    val = False
                else:
                    nothing
            self.o.set_config(se, val)

        self.o.add_readers_from_file(self.o.test_data_folder() +
            '../../opendrift/scripts/data_sources.txt')

        self.o.seed_cone(lon=lon, lat=lat, radius=radius,
                         time=start_time)#, #cone=cone,
                         #**extra_seed_args)

        time_step = self.o.get_config('general:time_step_minutes')*60
        time_step_output = self.o.get_config('general:time_step_output_minutes')*60
        duration = int(self.durationhours.get())*3600/time_step
        extra_args = {'time_step': time_step, 'time_step_output': time_step_output}
        if self.directionvar.get() == 'backwards':
            extra_args['time_step'] = -extra_args['time_step']
            extra_args['time_step_output'] = -extra_args['time_step_output']
        if self.has_diana is True:
            extra_args['outfile'] = self.outputdir + '/opendrift_' + \
                self.model.get() + self.o.start_time.strftime('_%Y%m%d_%H%M.nc')

        self.simulationname = 'opendrift_' + self.model.get() + \
            self.o.start_time.strftime('_%Y%m%d_%H%M')

        # Starting simulation run
        self.o.run(steps=duration, **extra_args)
        print(self.o)

        try:
            os.chmod(extra_args['outfile'], 0o666)
        except:
            pass

        # Model-specific post processing
        self.o.gui_postproc()


        self.results.destroy()
        self.results = tk.Frame(self.seed, bd=2,
                               relief=tk.FLAT, padx=5, pady=0)
        self.results.grid(row=70, column=3, columnspan=1, sticky='ew')
        tk.Button(self.results, text='Show animation',
                  command=lambda: self.handle_result(
                    'showanimation')).grid(row=10, column=1)
        tk.Button(self.results, text='Save animation',
                  command=lambda: self.handle_result(
                    'saveanimation')).grid(row=20, column=1)
        tk.Button(self.results, text='Show plot',
                  command=lambda: self.handle_result(
                    'showplot')).grid(row=30, column=1)
        tk.Button(self.results, text='Save plot',
                  command=lambda: self.handle_result(
                    'saveplot')).grid(row=40, column=1)
        if self.model.get() == 'OpenOil':
            tk.Button(self.results, text='Save oil budget',
                      command=lambda: self.handle_result(
                        'saveoilbudget')).grid(row=50, column=1)
            tk.Button(self.results, text='Show oil budget',
                      command=lambda: self.handle_result(
                        'showoilbudget')).grid(row=60, column=1)


        if self.model.get() =='RadionuclideDrift':
            tk.Button(self.results, text='Show animation specie',
                      command=lambda: self.handle_result(
                          'showanimationspecie')).grid(row=30, column=2)
            tk.Button(self.results, text='Save animation specie',
                      command=lambda: self.handle_result(
                          'saveanimationspecie')).grid(row=40, column=2)
            tk.Button(self.results, text='Animation profile',
                      command=lambda: self.handle_result(
                          'showanimationprofile')).grid(row=10, column=2)
#             tk.Button(self.results, text='Save conc file',
#                       command=lambda: self.handle_result(
#                           'saveconcfile')).grid(row=20, column=2)
            tk.Button(self.results, text='Plot conc',
                      command=lambda: self.handle_result(
                          'plotconc')).grid(row=20, column=2)



        if self.has_diana is True:
            diana_filename = self.dianadir + self.simulationname + '.nc'
            self.o.write_netcdf_density_map(diana_filename)
            tk.Button(self.results, text='Show in Diana',
                      command=lambda: os.system('diana &')
                      ).grid(row=80, column=1)

            try:
                os.chmod(diana_filename, 0o666)
            except:
                pass

def main():
    OpenDriftGUI().mainloop()

if __name__ == '__main__':
    OpenDriftGUI().mainloop()
